import zlib
import numba as nb
import numpy as np
from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.getmodel.manager import VulnerabilityIndex_dtype, VulnerabilityRow_dtype
from oasislmf.utils.exceptions import OasisException


@nb.njit(cache=True, error_model="numpy")
def _validate(data):
    if data.size == 0:
        return

    prev_vuln_id = data[0]["vulnerability_id"]
    prev_int_id = data[0]["intensity_bin_id"]
    prev_dmg_id = data[0]["damage_bin_id"]
    prob_sum = data[0]["probability"]

    # Check first damage_bin_id for each vuln_id/int_id starts at 1
    if prev_dmg_id != 1:
        raise OasisException(f"Error: First damage bin ID on line {0} is not 1 for vulnerability_id={prev_vuln_id}, intensity_bin_id={prev_int_id}.")

    next_expected_dmg_id = 2  # This is 2 as the first damage_bin_id must be 1

    for i in range(1, len(data)):
        row = data[i]
        v_id = row["vulnerability_id"]
        int_id = row["intensity_bin_id"]
        dmg_id = row["damage_bin_id"]
        prob = row["probability"]

        # Check vulnerability_id is non-decreasing
        if v_id < prev_vuln_id:
            raise OasisException(
                f"Error: Vulnerability IDs {prev_vuln_id} and {v_id} at line {i} are not in ascending order."
            )

        if v_id == prev_vuln_id and int_id == prev_int_id:
            # Check damage_bin_id is contiguous
            if dmg_id != next_expected_dmg_id:
                raise OasisException(f"Error: Non-contiguous damage bin IDs got {dmg_id}, expected {next_expected_dmg_id} on line {i}.")
            prob_sum += prob
            next_expected_dmg_id += 1
        else:
            # Check probabilities for each vuln_id/int_id group sum to 1
            if not np.isclose(prob_sum, 1, atol=1e-6):
                raise OasisException(
                    f"Error: Probabilities for vulnerability_id {prev_vuln_id} and intensity_bin_id {prev_int_id} do not sum to 1."
                    "total probability = {prob_sum}."
                )

            if v_id == prev_vuln_id:
                # Check intensity_bin_id is contiguous
                if int_id != prev_int_id + 1:
                    raise OasisException(
                        f"Error: Non contiguous intensity bin IDs, got {int_id}, expected {prev_int_id + 1} on line {i}."
                    )

            # Check first damage_bin_id for each vuln_id/int_id starts at 1
            if dmg_id != 1:
                raise OasisException(
                    f"Error: First damage bin ID on line {i} is not 1 for vulnerability_id={v_id}, intensity_bin_id={int_id}."
                )

            # Reset state
            prob_sum = prob
            next_expected_dmg_id = 2

        # Update prev ids
        prev_vuln_id = v_id
        prev_int_id = int_id
        prev_dmg_id = dmg_id

    # Check probabilities for each vuln_id/int_id group sum to 1
    if not np.isclose(prob_sum, 1, atol=1e-6):
        raise OasisException(
            f"Error: Probabilities for vulnerability_id {prev_vuln_id} and intensity_bin_id {prev_int_id} do not sum to 1."
            "total probability = {prob_sum}."
        )


def _validate_int_bins(data):
    from oasislmf.pytools.converters.csvtobin.manager import logger

    vuln_ids = np.unique(data['vulnerability_id'])
    int_ids_all = np.unique(data['intensity_bin_id'])

    missing_count = 0
    for v_id in vuln_ids:
        mask = data["vulnerability_id"] == v_id
        vuln_bins = np.unique(data["intensity_bin_id"][mask])
        missing_bins = np.setdiff1d(int_ids_all, vuln_bins)
        if missing_bins.size > 0:
            missing_count += 1
            logger.warning(
                f"WARNING: vulnerability_id {v_id} is missing intensity_bin_ids: {missing_bins.tolist()}"
            )
    if missing_count > 0:
        logger.warning("All intensity bins must be present for each vulnerability ID in single peril models.")


def vulnerability_tobin(
    stack,
    file_in,
    file_out,
    file_type,
    idx_file_out,
    max_damage_bin_idx,
    no_validation,
    suppress_int_bin_checks,
    zip_files
):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)

    if not no_validation:
        _validate(data)

    if not suppress_int_bin_checks:
        _validate_int_bins(data)

    # Write max_damage_bin to bin header
    np.array([max_damage_bin_idx], dtype=np.int32).tofile(file_out)

    if zip_files and idx_file_out is None:
        raise OasisException(f"Error: Cannot write zip files without provided idx_file_out zip path, currently {idx_file_out}")

    # Write straight to bin file as no idx file
    if idx_file_out is None:
        data.tofile(file_out)
        return

    idx_file_out = resolve_file(idx_file_out, "wb", stack)
    unique_vulns = np.unique(data["vulnerability_id"])
    offset = np.dtype(np.int32).itemsize  # Header offset

    for v_id in unique_vulns:
        vuln_mask = data["vulnerability_id"] == v_id
        vuln_data = data[vuln_mask]

        bin_data = np.empty(len(vuln_data), dtype=VulnerabilityRow_dtype)
        bin_data["intensity_bin_id"] = vuln_data["intensity_bin_id"]
        bin_data["damage_bin_id"] = vuln_data["damage_bin_id"]
        bin_data["probability"] = vuln_data["probability"]

        if any(bin_data["damage_bin_id"] > max_damage_bin_idx):
            raise OasisException(f"Error: Found damage_bin_id in data larger than max_damage_bin_idx: {max_damage_bin_idx}")

        bin_data = bin_data.tobytes()
        dsize = 0
        if zip_files:
            dsize = len(bin_data)
            bin_data = zlib.compress(bin_data)
        file_out.write(bin_data)
        size = len(bin_data)

        np.array([(v_id, offset, size, dsize)], dtype=VulnerabilityIndex_dtype).tofile(idx_file_out)
        offset += size



from pathlib import Path
import numpy as np
from oasislmf.pytools.common.data import write_ndarray_to_fmt_csv
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.pla.structure import read_lossfactors


def lossfactors_tocsv(stack, file_in, file_out, file_type, noheader):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    if str(file_in) == "-":
        plafactors = read_lossfactors(
            ignore_file_type=set(["csv"]),
            use_stdin=True
        )
    else:
        lossfactors_fp = Path(file_in)
        plafactors = read_lossfactors(
            run_dir=lossfactors_fp.parent,
            ignore_file_type=set(["csv"]),
            filename=lossfactors_fp.name
        )

    data = np.empty(len(plafactors), dtype=dtype)
    for i, (k, v) in enumerate(plafactors.items()):
        data[i] = (k[0], k[1], v)

    if not noheader:
        file_out.write(",".join(headers) + "\n")
    write_ndarray_to_fmt_csv(file_out, data, headers, fmt)



from pathlib import Path
import numpy as np
from oasislmf.pytools.common.data import write_ndarray_to_fmt_csv
from oasislmf.pytools.common.input_files import read_coverages
from oasislmf.pytools.converters.data import TOOL_INFO


def coverages_tocsv(stack, file_in, file_out, file_type, noheader):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    if str(file_in) == "-":
        coverages = read_coverages(
            ignore_file_type=set(["csv"]),
            use_stdin=True
        )
    else:
        cov_fp = Path(file_in)
        coverages = read_coverages(
            run_dir=cov_fp.parent,
            ignore_file_type=set(["csv"]),
            filename=cov_fp.name
        )
    data = np.zeros(len(coverages), dtype=dtype)
    data["coverage_id"] = np.arange(1, len(coverages) + 1)
    data["tiv"] = coverages

    if not noheader:
        file_out.write(",".join(headers) + "\n")
    write_ndarray_to_fmt_csv(file_out, data, headers, fmt)



#!/usr/bin/env python

import argparse
import logging

from oasislmf.pytools.eve.manager import DEFAULT_EVENTS_FILE
from . import logger, manager


def validate_flags(args):
    if args.process_number <= 0:
        raise RuntimeError(f"ERROR: evepy process_number {args.process_number} is not valid.")

    if args.total_processes <= 0:
        raise RuntimeError(f"ERROR: evepy total_processes {args.total_processes} is not valid.")

    if args.total_processes < args.process_number:
        raise ValueError(f"ERROR: evepy process_number {args.process_number} cannot be larger than the total_processes {args.total_processes}")


def main():
    parser = argparse.ArgumentParser(description='Generate partitioned event IDs stream.')
    parser.add_argument('process_number', type=int,
                        help='Process number to receive a partition of events. Processes are one-indexed.')
    parser.add_argument('total_processes', type=int,
                        help='Total number of processes to distribute the events over')

    parser.add_argument('-i', '--input_file', type=str, default=DEFAULT_EVENTS_FILE,
                        help=f'Input file containing the events binary (default: {DEFAULT_EVENTS_FILE})')
    parser.add_argument('-o', '--output_file', type=str, default='-',
                        help='Path to output file, if present then save output stream into output file instead of stdout.')

    parser.add_argument('-n', '--no_shuffle', action='store_true',
                        help='Disable shuffle. Events respect input ordering')
    parser.add_argument('-r', '--randomise', action='store_true',
                        help='Randomise events using the Fisher-Yates shuffle. Events are output during shuffling.')
    parser.add_argument('-R', '--randomise_builtin', action='store_true',
                        help='Randomise events using the builtin. Events are shuffled first and then output.')

    parser.add_argument('-v', '--logging-level', type=int, default=30,
                        help='logging level (debug:10, info:20, warning:30, error:40, critical:50)')

    args = parser.parse_args()
    validate_flags(args)
    kwargs = vars(args)

    # Set up logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    manager.main(**kwargs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
import logging

from . import manager, logger
from .data import VALID_EXT


def main():
    parser = argparse.ArgumentParser(description='Process average annual loss and standard deviation')
    parser.add_argument('--run_dir', type=str, default='.', help='path to the run directory')
    parser.add_argument('-K', '--subfolder', type=str, default=None, help='workspace sub folder name, inside <run_dir>/work/<sub folder name>')
    parser.add_argument('-a', '--aal', type=str, default=None, help='Output Average Annual Loss Table (AAL)')
    parser.add_argument('-c', '--alct', type=str, default=None, help='Output Average Loss Convergence Table (ALCT)')
    parser.add_argument('-M', '--meanonly', action='store_true', help='Output AAL with mean only')
    parser.add_argument('-l', '--confidence', type=float, default=0.95, help='0 <= confidence level <= 1, default 0.95')
    parser.add_argument('-v', '--logging-level', type=int, default=30,
                        help='logging level (debug:10, info:20, warning:30, error:40, critical:50)')
    parser.add_argument('-H', '--noheader', action='store_true', help='Suppress header in output files')
    parser.add_argument('-E', '--ext', type=str, default='csv', choices=VALID_EXT, help='Output data format')

    args = parser.parse_args()
    kwargs = vars(args)

    # Set up logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    manager.main(**kwargs)


if __name__ == '__main__':
    main()



from oasislmf.pytools.common.data import generate_output_metadata, oasis_int, oasis_float


VALID_EXT = ["csv", "bin", "parquet"]

SELT_output = [
    ('EventId', oasis_int, '%d'),
    ('SummaryId', oasis_int, '%d'),
    ('SampleId', oasis_int, '%d'),
    ('Loss', oasis_float, '%.2f'),
    ('ImpactedExposure', oasis_float, '%.2f'),
]

MELT_output = [
    ('EventId', oasis_int, '%d'),
    ('SummaryId', oasis_int, '%d'),
    ('SampleType', oasis_int, '%d'),
    ('EventRate', oasis_float, '%.6f'),
    ('ChanceOfLoss', oasis_float, '%.6f'),
    ('MeanLoss', oasis_float, '%.6f'),
    ('SDLoss', oasis_float, '%.6f'),
    ('MaxLoss', oasis_float, '%.6f'),
    ('FootprintExposure', oasis_float, '%.6f'),
    ('MeanImpactedExposure', oasis_float, '%.6f'),
    ('MaxImpactedExposure', oasis_float, '%.6f'),
]

QELT_output = [
    ('EventId', oasis_int, '%d'),
    ('SummaryId', oasis_int, '%d'),
    ('Quantile', oasis_float, '%.6f'),
    ('Loss', oasis_float, '%.6f'),
]

SELT_headers, SELT_dtype, SELT_fmt = generate_output_metadata(SELT_output)
MELT_headers, MELT_dtype, MELT_fmt = generate_output_metadata(MELT_output)
QELT_headers, QELT_dtype, QELT_fmt = generate_output_metadata(QELT_output)



"""
This file contains specific functionality needed for aggregate vulnerabilities.

"""
import logging

import numba as nb
import numpy as np
import pandas as pd
from numba import njit
from numba.typed import Dict, List
from numba.types import int32 as nb_int32

from oasis_data_manager.filestore.backends.base import BaseStorage
from oasislmf.pytools.common.data import areaperil_int, nb_areaperil_int, nb_oasis_float, aggregatevulnerability_dtype, vulnerability_weight_dtype

logger = logging.getLogger(__name__)

AGG_VULN_WEIGHTS_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int32))
AGG_VULN_WEIGHTS_VAL_TYPE = nb.types.float32


@njit(cache=True)
def gen_empty_agg_vuln_to_vuln_ids():
    """Generate empty map to store the definitions of aggregate vulnerability functions.

    Returns:
        dict[int, list[int]]: map of aggregate vulnerability id to list of vulnerability ids.
    """
    return Dict.empty(nb_int32, List.empty_list(nb_int32))


@njit(cache=True)
def gen_empty_areaperil_vuln_idx_to_weights():
    """Generate empty map to store the weights of individual vulnerability functions in each aggregate vulnerability.

    Returns:
        dict[AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE]: map of areaperil_id, vulnerability id to weight.
    """
    return Dict.empty(AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE)


def read_aggregate_vulnerability(storage: BaseStorage, ignore_file_type=set()):
    """Load the aggregate vulnerability definitions from file.

    Args:
        storage: (BaseStorage) the storage manager for fetching model data
        ignore_file_type (Set[str]): file extension to ignore when loading.

    Returns:
        np.array[AggregateVulnerability]: aggregate vulnerability table.
    """
    input_files = set(storage.listdir())

    if "aggregate_vulnerability.bin" in input_files and "bin" not in ignore_file_type:
        logger.debug(f"loading {storage.get_storage_url('aggregate_vulnerability.bin', encode_params=False)}")
        with storage.open('aggregate_vulnerability.bin') as f:
            aggregate_vulnerability = np.memmap(f, dtype=aggregatevulnerability_dtype, mode='r')

    elif "aggregate_vulnerability.csv" in input_files and "csv" not in ignore_file_type:
        logger.debug(f"loading {storage.get_storage_url('aggregate_vulnerability.csv', encode_params=False)}")
        with storage.open('aggregate_vulnerability.csv') as f:
            aggregate_vulnerability = np.loadtxt(f, dtype=aggregatevulnerability_dtype, delimiter=",", skiprows=1, ndmin=1)

    else:
        aggregate_vulnerability = None
        logging.warning(
            f"Aggregate vulnerability table not found at {storage.get_storage_url('', encode_params=False)[0]}. Continuing without aggregate vulnerability definitions.")

    return aggregate_vulnerability


def read_vulnerability_weights(storage: BaseStorage, ignore_file_type=set()):
    """Load the vulnerability weights definitions from file.

    Args:
        storage: (BaseStorage) the storage manager for fetching model data
        ignore_file_type (Set[str]): file extension to ignore when loading.

    Returns:
        np.array[VulnerabilityWeight]: vulnerability weights table.
    """
    input_files = set(storage.listdir())

    if "weights.bin" in input_files and "bin" not in ignore_file_type:
        logger.debug(f"loading {storage.get_storage_url('weights.bin', encode_params=False)}")
        with storage.open("weights.bin") as f:
            aggregate_weights = np.memmap(f, dtype=vulnerability_weight_dtype, mode='r')

    elif "weights.csv" in input_files and "csv" not in ignore_file_type:
        logger.debug(f"loading {storage.get_storage_url('weights.csv', encode_params=False)}")
        with storage.open("weights.csv") as f:
            aggregate_weights = np.loadtxt(f, dtype=vulnerability_weight_dtype, delimiter=",", skiprows=1, ndmin=1)

    else:
        aggregate_weights = None
        logging.warning(
            f"Vulnerability weights not found at {storage.get_storage_url('', encode_params=False)[0]}. Continuing without vulnerability weights definitions.")

    return aggregate_weights


def process_aggregate_vulnerability(aggregate_vulnerability):
    """Rearrange aggregate vulnerability definitions from tabular format to a map between aggregate
    vulnerability id and the list of vulnerability ids that it is made of.

    Args:
        aggregate_vulnerability (np.array[AggregateVulnerability]): aggregate vulnerability table.

    Returns:
        dict[int, list[int]]: map of aggregate vulnerability id to list of vulnerability ids.
    """
    agg_vuln_to_vuln_ids = gen_empty_agg_vuln_to_vuln_ids()

    if aggregate_vulnerability is not None:

        agg_vuln_df = pd.DataFrame(aggregate_vulnerability)
        # init agg_vuln_to_vuln_ids to allow numba to compile later functions
        # vulnerability_id and aggregate_vulnerability_id are remapped to the internal ids
        # using the vulnd_dict map that contains only the vulnerability_id used in this portfolio.

        # here we read all aggregate vulnerability_id, then, after processing the items file,
        # we will filter out the aggregate vulnerability that are not used in this portfolio.
        for agg, grp in agg_vuln_df.groupby('aggregate_vulnerability_id'):
            agg_vuln_id = nb_int32(agg)

            if agg_vuln_id not in agg_vuln_to_vuln_ids:
                agg_vuln_to_vuln_ids[agg_vuln_id] = List.empty_list(nb_int32)

            for entry in grp['vulnerability_id'].to_list():
                agg_vuln_to_vuln_ids[agg_vuln_id].append(nb_int32(entry))

    return agg_vuln_to_vuln_ids


@nb.njit(cache=True)
def process_vulnerability_weights(areaperil_vuln_i_to_weight, vuln_dict, aggregate_weights):
    """
    Polpulate the useful (areaperil_id, vulnerability_i) in areaperil_vuln_i_to_weight with the weight from aggregate_weights

    Args:
        areaperil_vuln_i_to_weight: dict of useful (areaperil_id, vulnerability_i) to 0. (weight placeholder to be updated)
        vuln_dict: vuln_dict (Tuple[Dict[int, int]): vulnerability dictionary, vuln_id => vuln_i.
        aggregate_weights (np.array[VulnerabilityWeight]): vulnerability weights table.
    """
    for i in range(len(aggregate_weights)):
        rec = aggregate_weights[i]
        if rec['vulnerability_id'] in vuln_dict:
            key = (nb_areaperil_int(rec['areaperil_id']), vuln_dict[rec['vulnerability_id']])
            if key in areaperil_vuln_i_to_weight:
                areaperil_vuln_i_to_weight[key] = nb_oasis_float(rec['weight'])



from oasislmf.pytools.common.data import fm_summary_xref_headers, fm_summary_xref_dtype, fm_summary_xref_fmt


headers = fm_summary_xref_headers
dtype = fm_summary_xref_dtype
fmt = fm_summary_xref_fmt
cli_support = ['bintocsv', 'csvtobin']



# lec/manager.py

import logging
import numpy as np
import numba as nb
from contextlib import ExitStack
from pathlib import Path
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from oasislmf.pytools.common.data import (DEFAULT_BUFFER_SIZE, oasis_int, oasis_float, oasis_int_size, oasis_float_size)
from oasislmf.pytools.common.event_stream import MAX_LOSS_IDX, MEAN_IDX, NUMBER_OF_AFFECTED_RISK_IDX, SUMMARY_STREAM_ID, init_streams_in, mv_read
from oasislmf.pytools.common.input_files import PERIODS_FILE, read_occurrence, read_periods, read_returnperiods
from oasislmf.pytools.lec.data import (AEP, AEPTVAR, AGG_FULL_UNCERTAINTY, AGG_SAMPLE_MEAN, AGG_WHEATSHEAF, AGG_WHEATSHEAF_MEAN,
                                       OCC_FULL_UNCERTAINTY, OCC_SAMPLE_MEAN, OCC_WHEATSHEAF, OCC_WHEATSHEAF_MEAN, OEP, OEPTVAR,
                                       OUTLOSS_DTYPE, EPT_dtype, EPT_fmt, EPT_headers, PSEPT_dtype, PSEPT_fmt, PSEPT_headers)
from oasislmf.pytools.lec.aggreports import AggReports
from oasislmf.pytools.lec.utils import get_outloss_mean_idx, get_outloss_sample_idx
from oasislmf.pytools.utils import redirect_logging


logger = logging.getLogger(__name__)


def read_input_files(
    run_dir,
    use_return_period,
    agg_wheatsheaf_mean,
    occ_wheatsheaf_mean,
    sample_size
):
    """Reads all input files and returns a dict of relevant data
    Args:
        run_dir (str | os.PathLike): Path to directory containing required files structure
        use_return_period (bool): Use Return Period file.
        agg_wheatsheaf_mean (bool): Aggregate Wheatsheaf Mean.
        occ_wheatsheaf_mean (bool): Occurrence Wheatsheaf Mean.
        sample_size (int): Sample Size.
    Returns:
        file_data (Dict[str, Any]): A dict of relevent data extracted from files
        use_return_period (bool): Use Return Period file.
        agg_wheatsheaf_mean (bool): Aggregate Wheatsheaf Mean.
        occ_wheatsheaf_mean (bool): Occurrence Wheatsheaf Mean.
    """
    input_dir = Path(run_dir, "input")
    occ_map, date_algorithm, granular_date, no_of_periods = read_occurrence(input_dir)
    period_weights = read_periods(no_of_periods, input_dir)
    periods_fp = Path(input_dir, PERIODS_FILE)
    if not periods_fp.exists():
        period_weights = np.array([], dtype=period_weights.dtype)
    else:  # Normalise period weights
        period_weights["weighting"] /= sample_size
    returnperiods, use_return_period = read_returnperiods(use_return_period, input_dir)

    # User must define return periods if he/she wishes to use non-uniform period weights for
    # Wheatsheaf/per sample mean output
    if agg_wheatsheaf_mean or occ_wheatsheaf_mean:
        if len(period_weights) > 0 and not use_return_period:
            logger.warning("WARNING: Return periods file must be present if you wish to use non-uniform period weights for Wheatsheaf mean/per sample mean output.")
            logger.info("INFO: Wheatsheaf mean/per sample mean output will not be produced.")
            agg_wheatsheaf_mean = False
            occ_wheatsheaf_mean = False
        elif len(period_weights) > 0:
            logger.info("INFO: Tail Value at Risk for Wheatsheaf mean/per sample mean is not supported if you wish to use non-uniform period weights.")

    file_data = {
        "occ_map": occ_map,
        "date_algorithm": date_algorithm,
        "granular_date": granular_date,
        "no_of_periods": no_of_periods,
        "period_weights": period_weights,
        "returnperiods": returnperiods,
    }
    return file_data, use_return_period, agg_wheatsheaf_mean, occ_wheatsheaf_mean


@nb.njit(cache=True, error_model="numpy")
def get_max_summary_id(file_handles):
    """Get max summary_id from all summary files
    Args:
        file_handles (List[np.memmap]): List of memmaps for summary files data
    Returns:
        max_summary_id (int): Max summary ID
    """
    max_summary_id = -1
    for fin in file_handles:
        cursor = oasis_int_size * 3

        valid_buff = len(fin)
        while cursor < valid_buff:
            _, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
            summary_id, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
            _, cursor = mv_read(fin, cursor, oasis_float, oasis_float_size)

            max_summary_id = max(max_summary_id, summary_id)

            while cursor < valid_buff:
                sidx, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
                _, cursor = mv_read(fin, cursor, oasis_float, oasis_float_size)
                if sidx == 0:
                    break
    return max_summary_id


@nb.njit(cache=True, fastmath=True, error_model="numpy")
def do_lec_output_agg_summary(
    summary_id,
    sidx,
    loss,
    filtered_occ_map,
    outloss_mean,
    outloss_sample,
    num_sidxs,
    max_summary_id,
):
    """Populate outloss_mean and outloss_sample with aggregate and max losses
    Args:
        summary_id (oasis_int): summary_id
        sidx (oasis_int): Sample ID
        loss (oasis_float): Loss value
        filtered_occ_map (ndarray[occ_map_dtype]): Filtered numpy map of event_id, period_no, occ_date_id from the occurrence file_
        outloss_mean (ndarray[OUTLOSS_DTYPE]): ndarray indexed by summary_id, period_no containing aggregate and max losses
        outloss_sample (ndarray[OUTLOSS_DTYPE]): ndarray indexed by summary_id, sidx, period_no containing aggregate and max losses
        num_sidxs (int): Number of sidxs to consider for outloss_sample
        max_summary_id (int): Max summary ID
    """
    for row in filtered_occ_map:
        period_no = row["period_no"]
        if sidx == MEAN_IDX:
            idx = get_outloss_mean_idx(period_no, summary_id, max_summary_id)
            outloss = outloss_mean
        else:
            idx = get_outloss_sample_idx(period_no, sidx, summary_id, num_sidxs, max_summary_id)
            outloss = outloss_sample
        outloss[idx]["row_used"] = True
        outloss[idx]["agg_out_loss"] += loss
        max_out_loss = max(outloss[idx]["max_out_loss"], loss)
        outloss[idx]["max_out_loss"] = max_out_loss


@nb.njit(cache=True, error_model="numpy")
def process_input_file(
    fin,
    outloss_mean,
    outloss_sample,
    summary_ids,
    occ_map,
    use_return_period,
    num_sidxs,
    max_summary_id,
):
    """Process summary file and populate outloss_mean and outloss_sample with losses
    Args:
        fin (np.memmap): summary binary memmap
        outloss_mean (ndarray[OUTLOSS_DTYPE]): ndarray indexed by summary_id, period_no containing aggregate and max losses
        outloss_sample (ndarray[OUTLOSS_DTYPE]): ndarray indexed by summary_id, sidx, period_no containing aggregate and max losses
        summary_ids (ndarray[bool]): bool array marking which summary_ids are used
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
        use_return_period (bool): Use Return Period file.
        num_sidxs (int): Number of sidxs to consider for outloss_sample
        max_summary_id (int): Max summary ID
    """
    # Set cursor to end of stream header (stream_type, sample_size, summary_set_id)
    cursor = oasis_int_size * 3

    # Read all samples
    valid_buff = len(fin)
    while cursor < valid_buff:
        event_id, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
        summary_id, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
        expval, cursor = mv_read(fin, cursor, oasis_float, oasis_float_size)

        # Discard samples if event_id not found
        if event_id not in occ_map:
            while cursor < valid_buff:
                sidx, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
                _, cursor = mv_read(fin, cursor, oasis_float, oasis_float_size)
                if sidx == 0:
                    break
            continue

        filtered_occ_map = occ_map[event_id]
        summary_ids[summary_id - 1] = True

        while cursor < valid_buff:
            sidx, cursor = mv_read(fin, cursor, oasis_int, oasis_int_size)
            loss, cursor = mv_read(fin, cursor, oasis_float, oasis_float_size)
            if sidx == 0:
                break
            if sidx == NUMBER_OF_AFFECTED_RISK_IDX or sidx == MAX_LOSS_IDX:
                continue
            if loss > 0 or use_return_period:
                do_lec_output_agg_summary(
                    summary_id,
                    sidx,
                    loss,
                    filtered_occ_map,
                    outloss_mean,
                    outloss_sample,
                    num_sidxs,
                    max_summary_id,
                )


@nb.njit(cache=True, error_model="numpy")
def run_lec(
    file_handles,
    outloss_mean,
    outloss_sample,
    summary_ids,
    occ_map,
    use_return_period,
    num_sidxs,
    max_summary_id,
):
    """Process each summary file and populate outloss_mean and outloss_sample
    Args:
        file_handles (List[np.memmap]): List of memmaps for summary files data
        outloss_mean (ndarray[OUTLOSS_DTYPE]): ndarray indexed by summary_id, period_no containing aggregate and max losses
        outloss_sample (ndarray[OUTLOSS_DTYPE]): ndarray indexed by summary_id, sidx, period_no containing aggregate and max losses
        summary_ids (ndarray[bool]): bool array marking which summary_ids are used
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
        use_return_period (bool): Use Return Period file.
        num_sidxs (int): Number of sidxs to consider for outloss_sample
        max_summary_id (int): Max summary ID
    """
    for fin in file_handles:
        process_input_file(
            fin,
            outloss_mean,
            outloss_sample,
            summary_ids,
            occ_map,
            use_return_period,
            num_sidxs,
            max_summary_id,
        )


def run(
    run_dir,
    subfolder,
    ept_output_file=None,
    psept_output_file=None,
    agg_full_uncertainty=False,
    agg_wheatsheaf=False,
    agg_sample_mean=False,
    agg_wheatsheaf_mean=False,
    occ_full_uncertainty=False,
    occ_wheatsheaf=False,
    occ_sample_mean=False,
    occ_wheatsheaf_mean=False,
    use_return_period=False,
    noheader=False,
    output_format="csv",
):
    """Runs LEC calculations
    Args:
        run_dir (str | os.PathLike): Path to directory containing required files structure
        subfolder (str): Workspace subfolder inside <run_dir>/work/<subfolder>
        ept_output_file (str, optional): Path to EPT output file. Defaults to None
        psept_output_file (str, optional): Path to PSEPT output file. Defaults to None
        agg_full_uncertainty (bool, optional): Aggregate Full Uncertainty. Defaults to False.
        agg_wheatsheaf (bool, optional): Aggregate Wheatsheaf. Defaults to False.
        agg_sample_mean (bool, optional): Aggregate Sample Mean. Defaults to False.
        agg_wheatsheaf_mean (bool, optional): Aggregate Wheatsheaf Mean. Defaults to False.
        occ_full_uncertainty (bool, optional): Occurrence Full Uncertainty. Defaults to False.
        occ_wheatsheaf (bool, optional): Occurrence Wheatsheaf. Defaults to False.
        occ_sample_mean (bool, optional): Occurrence Sample Mean. Defaults to False.
        occ_wheatsheaf_mean (bool, optional): Occurrence Wheatsheaf Mean. Defaults to False.
        use_return_period (bool, optional): Use Return Period file. Defaults to False.
        noheader (bool): Boolean value to skip header in output file
        output_format (str): Output format extension. Defaults to "csv".
    """
    outmap = {
        "ept": {
            "compute": ept_output_file is not None,
            "file_path": ept_output_file,
            "fmt": EPT_fmt,
            "headers": EPT_headers,
            "file": None,
            "dtype": EPT_dtype,
        },
        "psept": {
            "compute": psept_output_file is not None,
            "file_path": psept_output_file,
            "fmt": PSEPT_fmt,
            "headers": PSEPT_headers,
            "file": None,
            "dtype": PSEPT_dtype,
        },
    }

    output_format = "." + output_format
    output_binary = output_format == ".bin"
    output_parquet = output_format == ".parquet"
    # Check for correct suffix
    for path in [v["file_path"] for v in outmap.values()]:
        if path is None:
            continue
        if Path(path).suffix == "":  # Ignore suffix for pipes
            continue
        if (Path(path).suffix != output_format):
            raise ValueError(f"Invalid file extension for {output_format}, got {path},")

    if not all([v["compute"] for v in outmap.values()]):
        logger.warning("No output files specified")

    with ExitStack() as stack:
        workspace_folder = Path(run_dir, "work", subfolder)
        if not workspace_folder.is_dir():
            raise RuntimeError(f"Error: Unable to open directory {workspace_folder}")

        # work folder for lec files
        lec_files_folder = Path(workspace_folder, "lec_files")
        lec_files_folder.mkdir(parents=False, exist_ok=True)

        # Find summary binary files
        files = [file for file in workspace_folder.glob("*.bin")]
        file_handles = [np.memmap(file, mode="r", dtype="u1") for file in files]

        streams_in, (stream_source_type, stream_agg_type, sample_size) = init_streams_in(files, stack)
        if stream_source_type != SUMMARY_STREAM_ID:
            raise RuntimeError(f"Error: Not a summary stream type {stream_source_type}")

        max_summary_id = get_max_summary_id(file_handles)

        file_data, use_return_period, agg_wheatsheaf_mean, occ_wheatsheaf_mean = read_input_files(
            run_dir,
            use_return_period,
            agg_wheatsheaf_mean,
            occ_wheatsheaf_mean,
            sample_size,
        )

        output_flags = [
            agg_full_uncertainty,
            agg_wheatsheaf,
            agg_sample_mean,
            agg_wheatsheaf_mean,
            occ_full_uncertainty,
            occ_wheatsheaf,
            occ_sample_mean,
            occ_wheatsheaf_mean,
        ]

        # Check output_flags against output files
        handles_agg = [AGG_FULL_UNCERTAINTY, AGG_SAMPLE_MEAN, AGG_WHEATSHEAF_MEAN]
        handles_occ = [OCC_FULL_UNCERTAINTY, OCC_SAMPLE_MEAN, OCC_WHEATSHEAF_MEAN]
        handles_psept = [AGG_WHEATSHEAF, OCC_WHEATSHEAF]
        hasAGG = any([output_flags[idx] for idx in handles_agg])
        hasOCC = any([output_flags[idx] for idx in handles_occ])
        hasEPT = hasAGG or hasOCC
        hasPSEPT = any([output_flags[idx] for idx in handles_psept])
        outmap["ept"]["compute"] = outmap["ept"]["compute"] and hasEPT
        outmap["psept"]["compute"] = outmap["psept"]["compute"] and hasPSEPT

        if not outmap["ept"]["compute"]:
            logger.warning("WARNING: no valid output stream to fill EPT file")
        if not outmap["psept"]["compute"]:
            logger.warning("WARNING: no valid output stream to fill PSEPT file")
        if not (outmap["ept"]["compute"] or outmap["psept"]["compute"]):
            return

        # Create outloss array maps
        # outloss_mean has only -1 SIDX
        outloss_mean_file = Path(lec_files_folder, "lec_outloss_mean.bdat")
        outloss_mean = np.memmap(
            outloss_mean_file,
            dtype=OUTLOSS_DTYPE,
            mode="w+",
            shape=(file_data["no_of_periods"] * max_summary_id),
        )

        # outloss_sample has all SIDXs plus -2 and -3
        num_sidxs = sample_size + 2
        outloss_sample_file = Path(lec_files_folder, "lec_outloss_sample.bdat")
        outloss_sample = np.memmap(
            outloss_sample_file,
            dtype=OUTLOSS_DTYPE,
            mode="w+",
            shape=(file_data["no_of_periods"] * num_sidxs * max_summary_id),
        )

        # Array of Summary IDs found
        summary_ids = np.zeros((max_summary_id), dtype=np.bool_)

        # Run LEC calculations to populate outloss and summary_ids arrays
        run_lec(
            file_handles,
            outloss_mean,
            outloss_sample,
            summary_ids,
            file_data["occ_map"],
            use_return_period,
            num_sidxs,
            max_summary_id,
        )

        # Initialise output files LEC
        if output_binary:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                out_file = stack.enter_context(open(outmap[out_type]["file_path"], 'wb'))
                outmap[out_type]["file"] = out_file
        elif output_parquet:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                temp_out_data = np.zeros(DEFAULT_BUFFER_SIZE, dtype=outmap[out_type]["dtype"])
                temp_df = pd.DataFrame(temp_out_data, columns=outmap[out_type]["headers"])
                temp_table = pa.Table.from_pandas(temp_df)
                out_file = pq.ParquetWriter(outmap[out_type]["file_path"], temp_table.schema)
                outmap[out_type]["file"] = out_file
        else:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                out_file = stack.enter_context(open(outmap[out_type]["file_path"], 'w'))
                if not noheader:
                    csv_headers = ','.join(outmap[out_type]["headers"])
                    out_file.write(csv_headers + '\n')
                outmap[out_type]["file"] = out_file

        # Output aggregate reports to CSVs
        agg = AggReports(
            outmap,
            outloss_mean,
            outloss_sample,
            file_data["period_weights"],
            max_summary_id,
            sample_size,
            file_data["no_of_periods"],
            num_sidxs,
            use_return_period,
            file_data["returnperiods"],
            lec_files_folder,
            output_binary,
            output_parquet,
        )

        # Output Mean Damage Ratio
        if outmap["ept"]["compute"]:
            if hasOCC:
                agg.output_mean_damage_ratio(OEP, OEPTVAR, "max_out_loss")
            if hasAGG:
                agg.output_mean_damage_ratio(AEP, AEPTVAR, "agg_out_loss")

        # Output Full Uncertainty
        if output_flags[OCC_FULL_UNCERTAINTY]:
            agg.output_full_uncertainty(OEP, OEPTVAR, "max_out_loss")
        if output_flags[AGG_FULL_UNCERTAINTY]:
            agg.output_full_uncertainty(AEP, AEPTVAR, "agg_out_loss")

        # Output Wheatsheaf and Wheatsheaf Mean
        if output_flags[OCC_WHEATSHEAF] or output_flags[OCC_WHEATSHEAF_MEAN]:
            agg.output_wheatsheaf_and_wheatsheafmean(
                OEP, OEPTVAR, "max_out_loss",
                output_flags[OCC_WHEATSHEAF], output_flags[OCC_WHEATSHEAF_MEAN]
            )
        if output_flags[AGG_WHEATSHEAF] or output_flags[AGG_WHEATSHEAF_MEAN]:
            agg.output_wheatsheaf_and_wheatsheafmean(
                AEP, AEPTVAR, "agg_out_loss",
                output_flags[AGG_WHEATSHEAF], output_flags[AGG_WHEATSHEAF_MEAN]
            )

        # Output Sample Mean
        if output_flags[OCC_SAMPLE_MEAN]:
            agg.output_sample_mean(OEP, OEPTVAR, "max_out_loss")
        if output_flags[AGG_SAMPLE_MEAN]:
            agg.output_sample_mean(AEP, AEPTVAR, "agg_out_loss")


@redirect_logging(exec_name='lecpy')
def main(
    run_dir='.',
    subfolder=None,
    ept=None,
    psept=None,
    agg_full_uncertainty=False,
    agg_wheatsheaf=False,
    agg_sample_mean=False,
    agg_wheatsheaf_mean=False,
    occ_full_uncertainty=False,
    occ_wheatsheaf=False,
    occ_sample_mean=False,
    occ_wheatsheaf_mean=False,
    use_return_period=False,
    noheader=False,
    ext="csv",
    **kwargs
):
    run(
        run_dir,
        subfolder,
        ept_output_file=ept,
        psept_output_file=psept,
        agg_full_uncertainty=agg_full_uncertainty,
        agg_wheatsheaf=agg_wheatsheaf,
        agg_sample_mean=agg_sample_mean,
        agg_wheatsheaf_mean=agg_wheatsheaf_mean,
        occ_full_uncertainty=occ_full_uncertainty,
        occ_wheatsheaf=occ_wheatsheaf,
        occ_sample_mean=occ_sample_mean,
        occ_wheatsheaf_mean=occ_wheatsheaf_mean,
        use_return_period=use_return_period,
        noheader=noheader,
        output_format=ext,
    )



"""
Json to Markdown Utils
"""
__all__ = [
    'BaseJsonToMarkdownGenerator',
    'DefaultJsonToMarkdownGenerator',
    'RDLS_0_2_0_JsonToMarkdownGenerator',
]


from .base import BaseJsonToMarkdownGenerator
from .default import DefaultJsonToMarkdownGenerator
from .rdls_0__2__0 import RDLS_0_2_0_JsonToMarkdownGenerator



__all__ = [
    'APIClient',
    'ApiEndpoint',
    'FileEndpoint',
    'JsonEndpoint',
    'API_analyses',
    'API_models',
    'API_portfolios',
]

import io
import json
import logging
import os
import tarfile
import time
import pathlib

from posixpath import join as urljoin

import pandas as pd

from tqdm import tqdm
from requests_toolbelt import MultipartEncoder
from requests.exceptions import (
    HTTPError,
)
from .session import APISession
from ..utils.exceptions import OasisException


class ApiEndpoint(object):
    """
    Used to Implement the default requests common to all Oasis API
    End points.
    """

    def __init__(self, session, url_endpoint, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.session = session
        self.url_endpoint = str(url_endpoint)

    def create(self, data):
        return self.session.post(self.url_endpoint, json=data)

    def get(self, ID=None):
        if ID:
            return self.session.get(urljoin(self.url_endpoint, f'{ID}/'))
        return self.session.get(self.url_endpoint)

    def delete(self, ID):
        return self.session.delete(urljoin(self.url_endpoint, f'{ID}/'))

    def search(self, metadata={}):
        search_string = ""
        for key in metadata:
            if not search_string:
                search_string = f'?{key}={metadata[key]}'
            else:
                search_string += f'&{key}={metadata[key]}'
        return self.session.get(f'{self.url_endpoint}{search_string}')


class JsonEndpoint(object):
    """
    Used for JSON data End points.
    """

    def __init__(self, session, url_endpoint, url_resource, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.session = session
        self.url_endpoint = str(url_endpoint)
        self.url_resource = str(url_resource)

    def _build_url(self, ID):
        return urljoin(self.url_endpoint, str(ID), self.url_resource)

    def get(self, ID):
        return self.session.get(self._build_url(ID))

    def post(self, ID, data):
        return self.session.post(self._build_url(ID), json=data)

    def delete(self, ID):
        return self.session.delete(self._build_url(ID))

    def download(self, ID, file_path, overwrite=True):
        abs_fp = os.path.realpath(os.path.expanduser(file_path))
        dir_fp = os.path.dirname(abs_fp)

        # Check and create base dir
        if not os.path.exists(dir_fp):
            os.makedirs(dir_fp)

        # Check if file exists
        if os.path.exists(abs_fp) and not overwrite:
            error_message = 'Local file alreday exists: {}'.format(abs_fp)
            raise IOError(error_message)

        with io.open(abs_fp, 'w', encoding='utf-8') as f:
            r = self.get(ID)
            f.write(json.dumps(r.json(), ensure_ascii=False, indent=4))
        return r


class FileEndpoint(object):
    """
    File Resources Endpoint for Upload / Downloading
    """

    def __init__(self, session, url_endpoint, url_resource, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.session = session
        self.url_endpoint = str(url_endpoint)
        self.url_resource = str(url_resource)

    def _build_url(self, ID):
        return urljoin(self.url_endpoint, str(ID), self.url_resource)

    def _set_content_type(self, file_path):
        content_type_map = {
            'parquet': 'application/octet-stream',
            'pq': 'application/octet-stream',
            'csv': 'text/csv',
            'gz': 'application/gzip',
            'zip': 'application/zip',
            'bz2': 'application/x-bzip2',
        }
        file_ext = pathlib.Path(file_path).suffix[1:].lower()
        return content_type_map[file_ext] if file_ext in content_type_map else 'text/csv'

    def upload(self, ID, file_path, content_type=None):
        if not content_type:
            content_type = self._set_content_type(file_path)
        return self.session.upload(self._build_url(ID), file_path, content_type)

    def upload_byte(self, ID, file_bytes, filename, content_type=None):
        if not content_type:
            content_type = self._set_content_type(filename)
        return self.session.upload_byte(self._build_url(ID), file_bytes, filename, content_type)

    def download(self, ID, file_path, overwrite=True, chuck_size=1024):
        abs_fp = os.path.realpath(os.path.expanduser(file_path))
        dir_fp = os.path.dirname(abs_fp)

        # Check and create base dir
        if not os.path.exists(dir_fp):
            os.makedirs(dir_fp)

        # Check if file exists
        if os.path.exists(abs_fp) and not overwrite:
            error_message = 'Local file alreday exists: {}'.format(abs_fp)
            raise IOError(error_message)

        with io.open(abs_fp, 'wb') as f:
            r = self.session.get(self._build_url(ID), stream=True)
            for chunk in r.iter_content(chunk_size=chuck_size):
                f.write(chunk)
            return r

    def get(self, ID):
        return self.session.get(self._build_url(ID))

    def get_dataframe(self, ID):
        '''
        Return file endpoint as dict of pandas Dataframes:

        either 'application/gzip': search and extract all csv
        or 'text/csv': return as dataframe
        '''
        supported_content = [
            'text/csv',
            'application/gzip',
            'application/octet-stream',
        ]
        r = self.get(ID)
        file_type = r.headers['Content-Type']

        if file_type not in supported_content:
            raise OasisException(f'Unsupported filetype for Dataframe conversion: {file_type}')

        if file_type == 'text/csv':
            return pd.read_csv(io.StringIO(r.content.decode('utf-8')))

        if file_type == 'application/octet-stream':
            return pd.read_parquet(io.BytesIO(r.content))

        if file_type == 'application/gzip':
            dataframes_list = {}
            tar = tarfile.open(fileobj=io.BytesIO(r.content))

            for member in [f for f in tar.getmembers() if '.csv' in f.name]:
                csv = tar.extractfile(member)
                dataframes_list[os.path.basename(member.name)] = pd.read_csv(csv)

            for member in [f for f in tar.getmembers() if '.parquet' in f.name]:
                pq = tar.extractfile(member)
                dataframes_list[os.path.basename(member.name)] = pd.read_parquet(pq)
            return dataframes_list

    def post(self, ID, data_object, content_type='application/json'):
        m = MultipartEncoder(fields={'file': ('data', data_object, content_type)})
        return self.session.post(
            self._build_url(ID),
            data=m,
            headers={'Content-Type': m.content_type}
        )

    def post_dataframe(self, ID, data_frame):
        csv_buffer = io.StringIO()
        data_frame.to_csv(csv_buffer, index=False)
        return self.post(ID, data_object=csv_buffer, content_type='text/csv')

    def delete(self, ID):
        return self.session.delete(self._build_url(ID))


class SettingTemplatesBaseEndpoint(object):
    def __init__(self, session, url_endpoint, url_resource=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.session = session
        self.url_endpoint = str(url_endpoint)
        self.url_resource = url_resource

    def _build_url(self, model_pk, ID=None):
        url_components = [self.url_endpoint, str(model_pk), 'setting_templates/']
        if ID is not None:
            url_components += [str(ID)]

        if self.url_resource is not None:
            url_components += [str(self.url_resource)]

        return urljoin(*url_components)

    def get(self, model_pk, ID=None):
        return self.session.get(self._build_url(model_pk, ID))

    def post(self, model_pk, data, ID=None):
        return self.session.post(self._build_url(model_pk, ID), json=data)

    def delete(self, model_pk, ID):
        return self.session.delete(self._build_url(model_pk, ID))

    def put(self, model_pk, ID, data):
        return self.session.put(self._build_url(model_pk, ID), json=data)

    def patch(self, model_pk, ID, data):
        return self.session.patch(self._build_url(model_pk, ID), json=data)


class SettingTemplatesEndpoint(SettingTemplatesBaseEndpoint):
    """
    Settings Template Endpoint for interacting with analysis settings templates for a given model.
    """

    def __init__(self, session, url_endpoint, logger=None):
        super().__init__(session, url_endpoint)
        self.content = SettingTemplatesBaseEndpoint(session, url_endpoint, 'content/', logger=logger)


class API_models(ApiEndpoint):
    def __init__(self, session, url_endpoint):
        super(API_models, self).__init__(session, url_endpoint)
        self.resource_file = FileEndpoint(self.session, self.url_endpoint, 'resource_file/')
        self.settings = JsonEndpoint(self.session, self.url_endpoint, 'settings/')
        self.versions = JsonEndpoint(self.session, self.url_endpoint, 'versions/')

        # Platform 2.0 only (Check might be needed here)
        self.chunking_configuration = JsonEndpoint(self.session, self.url_endpoint, 'chunking_configuration/')
        self.scaling_configuration = JsonEndpoint(self.session, self.url_endpoint, 'scaling_configuration/')

        self.setting_templates = SettingTemplatesEndpoint(self.session, self.url_endpoint)

    def data_files(self, ID):
        return self.session.get('{}{}/data_files'.format(self.url_endpoint, ID))

    def create(self, supplier_id, model_id, version_id, data_files=[]):
        data = {"supplier_id": supplier_id,
                "model_id": model_id,
                "version_id": version_id,
                "data_files": data_files}
        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, supplier_id, model_id, version_id, data_files=[]):
        data = {"supplier_id": supplier_id,
                "model_id": model_id,
                "version_id": version_id,
                "data_files": data_files}
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)


class API_portfolios(ApiEndpoint):

    def __init__(self, session, url_endpoint):
        super(API_portfolios, self).__init__(session, url_endpoint)
        self.accounts_file = FileEndpoint(self.session, self.url_endpoint, 'accounts_file/')
        self.location_file = FileEndpoint(self.session, self.url_endpoint, 'location_file/')
        self.reinsurance_info_file = FileEndpoint(self.session, self.url_endpoint, 'reinsurance_info_file/')
        self.reinsurance_scope_file = FileEndpoint(self.session, self.url_endpoint, 'reinsurance_scope_file/')
        self.storage_links = JsonEndpoint(self.session, self.url_endpoint, 'storage_links/')

    def create(self, name):
        data = {"name": name}
        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, name):
        data = {"name": name}
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)

    def create_analyses(self, ID, name, model_id):
        """ Create new analyses from Exisiting portfolio
        """
        data = {"name": name,
                "model": model_id}
        return self.session.post('{}{}/create_analysis/'.format(self.url_endpoint, ID), json=data)


class API_datafiles(ApiEndpoint):

    def __init__(self, session, url_endpoint):
        super(API_datafiles, self).__init__(session, url_endpoint)
        self.content = FileEndpoint(self.session, self.url_endpoint, 'content/')

    def create(self, file_description, file_category=None):
        data = {"file_description": file_description}
        if file_category is not None:
            data["file_category"] = file_category

        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, file_description, file_category=None):
        data = {"file_description": file_description}
        if file_category is not None:
            data["file_category"] = file_category
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)


class API_task_status(ApiEndpoint):

    def __init__(self, session, url_endpoint):
        super(API_task_status, self).__init__(session, url_endpoint)
        self.output_log = FileEndpoint(self.session, self.url_endpoint, 'output_log/')
        self.error_log = FileEndpoint(self.session, self.url_endpoint, 'error_log/')


class API_analyses(ApiEndpoint):

    def __init__(self, session, url_endpoint):
        super(API_analyses, self).__init__(session, url_endpoint)
        self.lookup_errors_file = FileEndpoint(self.session, self.url_endpoint, 'lookup_errors_file/')
        self.lookup_success_file = FileEndpoint(self.session, self.url_endpoint, 'lookup_success_file/')
        self.lookup_validation_file = FileEndpoint(self.session, self.url_endpoint, 'lookup_validation_file/')
        self.summary_levels_file = FileEndpoint(self.session, self.url_endpoint, 'summary_levels_file/')
        self.input_file = FileEndpoint(self.session, self.url_endpoint, 'input_file/')
        self.input_generation_traceback_file = FileEndpoint(self.session, self.url_endpoint, 'input_generation_traceback_file/')
        self.output_file = FileEndpoint(self.session, self.url_endpoint, 'output_file/')
        self.run_traceback_file = FileEndpoint(self.session, self.url_endpoint, 'run_traceback_file/')
        self.run_log_file = FileEndpoint(self.session, self.url_endpoint, 'run_log_file/')
        self.settings_file = FileEndpoint(self.session, self.url_endpoint, 'settings_file/')
        self.settings = JsonEndpoint(self.session, self.url_endpoint, 'settings/')

    def create(self, name, portfolio_id, model_id, data_files=[]):
        data = {"name": name,
                "portfolio": portfolio_id,
                "model": model_id,
                "complex_model_data_files": data_files}
        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, name, portfolio_id, model_id, data_files=[]):
        data = {"name": name,
                "portfolio": portfolio_id,
                "model": model_id,
                "complex_model_data_files": data_files}
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)

    def status(self, ID):
        return self.get(ID).json()['status']

    def generate(self, ID):
        return self.session.post('{}{}/generate_inputs/'.format(self.url_endpoint, ID), json={})

    def run(self, ID):
        return self.session.post('{}{}/run/'.format(self.url_endpoint, ID), json={})

    def generate_and_run(self, ID):
        return self.session.post('{}{}/generate_and_run/'.format(self.url_endpoint, ID), json={})

    def cancel_analysis_run(self, ID):
        return self.session.post('{}{}/cancel_analysis_run/'.format(self.url_endpoint, ID), json={})

    def cancel_generate_inputs(self, ID):
        return self.session.post('{}{}/cancel_generate_inputs/'.format(self.url_endpoint, ID), json={})

    def cancel(self, ID):
        return self.session.post('{}{}/cancel/'.format(self.url_endpoint, ID), json={})

    def copy(self, ID):
        return self.session.post('{}{}/copy/'.format(self.url_endpoint, ID), json={})

    def data_files(self, ID):
        return self.session.get('{}{}/data_files/'.format(self.url_endpoint, ID))

    def storage_links(self, ID):
        return self.session.get('{}{}/storage_links/'.format(self.url_endpoint, ID))

    def sub_task_list(self, ID):
        return self.session.get('{}{}/sub_task_list/'.format(self.url_endpoint, ID))

# --- API Main Client ------------------------------------------------------- #


class APIClient(object):
    def __init__(self, api_url='http://localhost:8000', api_ver='V2', username='admin', password='password', timeout=25, logger=None, **kwargs):
        self.logger = logger or logging.getLogger(__name__)

        self.api = APISession(api_url, username, password, timeout, **kwargs)
        self.api_ver = api_ver.lower()
        self.models = API_models(self.api, '{}{}/models/'.format(self.api.url_base, self.api_ver))
        self.portfolios = API_portfolios(self.api, '{}{}/portfolios/'.format(self.api.url_base, self.api_ver))
        self.analyses = API_analyses(self.api, '{}{}/analyses/'.format(self.api.url_base, self.api_ver))
        self.data_files = API_datafiles(self.api, '{}{}/data_files/'.format(self.api.url_base, self.api_ver))
        self.task_status = API_task_status(self.api, '{}{}/analysis-task-statuses/'.format(self.api.url_base, self.api_ver))

    def oed_peril_codes(self):
        return self.api.get('{}oed_peril_codes/'.format(self.api.url_base))

    def server_info(self):
        return self.api.get('{}server_info/'.format(self.api.url_base))

    def healthcheck(self):
        return self.api.get('{}healthcheck/'.format(self.api.url_base))

    def upload_portfolio_file(self, portfolio_id, portfolio_file, upload_data):
        """
        Upload a portfolio file using the API. Supports absolute filepaths or
        bytestreams.

        If uploading a byte stream `upload_data` is a dict with the keys `name`
        which is a `str` containing the filename and `bytes` containing the
        byte stream of the file data. For example:

        ```python
        upload_portfolio_file(<portfolio_id>, "location_file",
                              {'bytes': <byte_stream>, 'name': <filename>})
        ```

        Parameters
        :param portfolio_id: Portfolio {id} from
        :type portfolio_id: int

        :param portfolio_file: The name of the portfolio file to update. One of
            the following options `location_file`, `accounts_file`,
            `reinsurance_info_file` or `reinsurance_scope_file`.
        :type settings: str

        :param upload_data: The file to upload through the api. This should be
        a `str` if it is a filepath or a `dict` if it is a byte stream with the
        keys `name` and `bytes` corresponding to the filename and bytestream
        respectively.
        :type upload_data: [str, dict]
        ----------
        """
        if isinstance(upload_data, dict):
            getattr(self.portfolios, portfolio_file).upload_byte(portfolio_id,
                                                                 upload_data['bytes'],
                                                                 upload_data['name'])
            self.logger.info("File uploaded: {}".format(upload_data['name']))
        else:
            getattr(self.portfolios, portfolio_file).upload(portfolio_id, upload_data)
            self.logger.info("File uploaded: {}".format(upload_data))

    def upload_inputs(self, portfolio_name=None, portfolio_id=None,
                      location_fp=None, accounts_fp=None, ri_info_fp=None, ri_scope_fp=None):
        if not portfolio_name:
            portfolio_name = time.strftime("Portfolio_%d%m%Y-%H%M%S")

        try:
            if portfolio_id:
                self.logger.info('Updating exisiting portfolio')
                portfolio = self.portfolios.update(portfolio_id, portfolio_name)
            else:
                self.logger.info('Creating portfolio')
                portfolio = self.portfolios.create(portfolio_name)
                portfolio_id = portfolio.json()['id']

            # Upload exposure
            if location_fp:
                self.upload_portfolio_file(portfolio_id, 'location_file', location_fp)
            if accounts_fp:
                self.upload_portfolio_file(portfolio_id, 'accounts_file', accounts_fp)
            if ri_info_fp:
                self.upload_portfolio_file(portfolio_id, 'reinsurance_info_file', ri_info_fp)
            if ri_scope_fp:
                self.upload_portfolio_file(portfolio_id, 'reinsurance_scope_file', ri_scope_fp)
            return portfolio.json()
        except HTTPError as e:
            self.api.unrecoverable_error(e, 'upload_inputs: failed')

    def upload_settings(self, analyses_id, settings):
        """
        Upload an analyses run settings to an API

        Method to post JSON data or upload a settings file containing JSON data

        Parameters
        ----------
        :param analyses_id: Analyses settings {id} from, `v1/analyses/{id}/settings`
        :type analyses_id: int

        :param settings: Either a valid filepath or dictionary holding the settings
        :type settings: [str, dict]

        :return:
        :rtype None
        """
        if isinstance(settings, dict):
            self.analyses.settings.post(analyses_id, settings)
            self.logger.info("Settings JSON uploaded: {}".format(settings))

        elif os.path.isfile(str(settings)):
            with io.open(settings, 'r', encoding='utf-8') as f:
                data = json.load(f)
            self.analyses.settings.post(analyses_id, data)
            self.logger.info("Settings JSON uploaded: {}".format(settings))
        else:
            raise TypeError("'settings': not a valid filepath or dictionary")

    def create_analysis(self, portfolio_id, model_id, analysis_name=None, analysis_settings_fp=None):
        try:
            if not analysis_name:
                analysis_name = time.strftime("Analysis_%d%m%Y-%H%M%S")

            analyses = self.analyses.create(analysis_name, portfolio_id, model_id).json()
            if analysis_settings_fp:
                self.upload_settings(analyses['id'], analysis_settings_fp)

            return analyses
        except HTTPError as e:
            self.api.unrecoverable_error(e, 'create_analysis: failed')

    def run_generate(self, analysis_id, poll_interval=5):
        """
        Generates the inputs for the analysis based on the portfolio.
        The analysis must have one of the following statuses, `NEW`, `INPUTS_GENERATION_ERROR`,
        `INPUTS_GENERATION_CANCELLED`, `READY`, `RUN_COMPLETED`, `RUN_CANCELLED` or
        `RUN_ERROR`.
        """

        try:
            r = self.analyses.generate(analysis_id)
            analysis = r.json()
            self.logger.info('Inputs Generation: Starting (id={})'.format(analysis_id))
            logged_queued = None
            logged_running = None

            while True:
                if analysis['status'] in ['READY']:
                    self.logger.info('Inputs Generation: Complete (id={})'.format(analysis_id))
                    return True

                elif analysis['status'] in ['INPUTS_GENERATION_CANCELLED']:
                    self.logger.info('Input Generation: Cancelled (id={})'.format(analysis_id))
                    return False

                elif analysis['status'] in ['INPUTS_GENERATION_ERROR']:
                    self.logger.info('Input Generation: Failed (id={})'.format(analysis_id))
                    error_trace = self.analyses.input_generation_traceback_file.get(analysis_id).text
                    self.logger.error("\nServer logs:")
                    self.logger.error(error_trace)
                    return False

                elif analysis['status'] in ['INPUTS_GENERATION_QUEUED']:
                    if not logged_queued:
                        logged_queued = True
                        self.logger.info('Input Generation: Queued (id={})'.format(analysis_id))

                    time.sleep(poll_interval)
                    r = self.analyses.get(analysis_id)
                    analysis = r.json()
                    continue

                elif analysis['status'] in ['INPUTS_GENERATION_STARTED']:
                    if not logged_running:
                        logged_running = True
                        self.logger.info('Input Generation: Executing (id={})'.format(analysis_id))

                    if analysis.get('run_mode', '') == 'V2':
                        sub_tasks_list = self.analyses.sub_task_list(analysis_id).json()
                        with tqdm(total=len(sub_tasks_list),
                                  unit=' sub_task',
                                  desc='Input Generation') as pbar:

                            completed = []
                            while len(completed) < len(sub_tasks_list):
                                sub_tasks_list = self.analyses.sub_task_list(analysis_id).json()
                                analysis = self.analyses.get(analysis_id).json()
                                completed = [tsk for tsk in sub_tasks_list if tsk['status'] == 'COMPLETED']
                                pbar.update(len(completed) - pbar.n)
                                time.sleep(poll_interval)

                                # Exit conditions
                                if ('_CANCELLED' in analysis['status']) or ('_ERROR' in analysis['status']):
                                    break
                                elif 'READY' in analysis['status']:
                                    pbar.update(pbar.total - pbar.n)
                                    break

                    else:
                        time.sleep(poll_interval)
                        analysis = self.analyses.get(analysis_id).json()

                    continue

                else:
                    err_msg = "Input Generation: Unknown State'{}'".format(analysis['status'])
                    raise OasisException(err_msg)
        except HTTPError as e:
            self.api.unrecoverable_error(e, 'run_generate: failed')

    def run_analysis(self, analysis_id, analysis_settings_fp=None, poll_interval=5):
        """
        Runs all the analysis. The analysis must have one of the following
        statuses, `NEW`, `RUN_COMPLETED`, `RUN_CANCELLED` or
        `RUN_ERROR`
        """
        try:
            if analysis_settings_fp:
                self.upload_settings(analysis_id, analysis_settings_fp)

            self.analyses.run(analysis_id)
            self.logger.info(f'Analysis Run: Starting (id={analysis_id})')
            return self._poll_analysis_until_complete(analysis_id, poll_interval)

        except HTTPError as e:
            self.api.unrecoverable_error(e, 'run_analysis: failed')

    def _poll_analysis_until_complete(self, analysis_id, poll_interval):
        logged_queued = False
        while True:
            analysis = self.analyses.get(analysis_id).json()
            status = analysis['status']
            if status == 'RUN_COMPLETED':
                self.logger.info(f'Analysis Run: Complete (id={analysis_id})')
                return True
            elif status == 'RUN_CANCELLED':
                self.logger.info(f'Analysis Run: Cancelled (id={analysis_id})')
                return False
            elif status == 'RUN_ERROR':
                error_trace = self.analyses.run_traceback_file.get(analysis_id).text
                self.logger.error(f'Analysis Run: Failed (id={analysis_id})\n\nServer logs:\n{error_trace}')
                return False
            elif status == 'RUN_QUEUED':
                if not logged_queued:
                    self.logger.info(f'Analysis Run: Queued (id={analysis_id})')
                    logged_queued = True
            elif status == 'RUN_STARTED':
                self.logger.info(f'Analysis Run: Executing (id={analysis_id})')
                self._run_until_complete(analysis_id, poll_interval)
                poll_interval = 0
            else:
                err_msg = f"Execution status in Unknown State: '{analysis['status']}'"
                self.logger.error(err_msg)
                raise OasisException(err_msg)
            time.sleep(poll_interval)

    def _run_until_complete(self, analysis_id, poll_interval):
        analysis = self.analyses.get(analysis_id).json()
        while analysis.get('num_events_total', 0) == 0:
            # Possible to stay here if no env vars given to worker
            if any(status in analysis['status'] for status in ['_CANCELLED', '_ERROR', 'COMPLETED']):
                return
            time.sleep(poll_interval)
            analysis = self.analyses.get(analysis_id).json()

        with tqdm(total=analysis['num_events_total'], unit=' gul stages', desc='Analysis Run') as pbar:
            while "COMPLETED" not in analysis['status']:
                pbar.update(analysis['num_events_complete'] - pbar.n)
                time.sleep(poll_interval)

                analysis = self.analyses.get(analysis_id).json()
                if ('_CANCELLED' in analysis['status']) or ('_ERROR' in analysis['status']):
                    break
                elif 'COMPLETED' in analysis['status']:
                    pbar.update(pbar.total - pbar.n)
                    break

    def download_output(self, analysis_id, download_path='', filename=None, clean_up=False, overwrite=True):
        if not filename:
            filename = 'analysis_{}_output.tar.gz'.format(analysis_id)
        try:
            output_file = os.path.join(download_path, filename)
            self.analyses.output_file.download(ID=analysis_id, file_path=output_file, overwrite=overwrite)
            self.logger.info('Analysis Download output: filename={}, (id={})'.format(output_file, analysis_id))
            if clean_up:
                self.analyses.delete(analysis_id)
        except HTTPError as e:
            err_msg = 'Analysis Download output: Failed (id={})'.format(analysis_id)
            self.api.unrecoverable_error(e, err_msg)

    def cancel_generate(self, analysis_id):
        """
        Cancels a currently inputs generation. The analysis status must be `GENERATING_INPUTS`
        """
        try:
            self.analyses.cancel_generate_inputs(analysis_id)
            self.logger.info('Cancelled Input generation: (Id={})'.format(analysis_id))
        except HTTPError as e:
            self.api.unrecoverable_error(e, 'cancel_generate: Failed')

    def cancel_analysis(self, analysis_id):
        """
        Cancels a currently running analysis. The analysis must have one of the following
        statuses, `PENDING` or `STARTED`
        """
        try:
            self.analyses.cancel_analysis_run(analysis_id)
            self.logger.info('Cancelled analysis run: (Id={})'.format(analysis_id))
        except HTTPError as e:
            self.api.unrecoverable_error(e, 'cancel_analysis: Failed')



# --- Ktools constants --------------------------------------------------------- #
DEDUCTIBLE_AND_LIMIT_CALCRULE_ID = 1
FRANCHISE_DEDUCTIBLE_AND_LIMIT_CALCRULE_ID = 3
DEDUCTIBLE_ONLY_CALCRULE_ID = 12
DEDUCTIBLE_AS_A_CAP_ON_THE_RETENTION_OF_INPUT_LOSSES_CALCRULE_ID = 10
DEDUCTIBLE_AS_A_FLOOR_ON_THE_RETENTION_OF_INPUT_LOSSES_CALCRULE_ID = 11
DEDUCTIBLE_LIMIT_AND_SHARE_CALCRULE_ID = 2
DEDUCTIBLE_AND_LIMIT_AS_A_PROPORTION_OF_LOSS_CALCRUKE_ID = 5
DEDUCTIBLE_WITH_LIMIT_AS_A_PROPORTION_OF_LOSS_CALCRUKE_ID = 9
LIMIT_ONLY_CALCRULE_ID = 14
LIMIT_AS_A_PROPORTION_OF_LOSS_CALCRULE_ID = 15
DEDUCTIBLE_AS_A_PROPORTION_OF_LOSS_CALCRULE_ID = 16

CALCRULE_ID_DEDUCTIBLE_AND_LIMIT = 1
CALCRULE_ID_DEDUCTIBLE_ATTACHMENT_LIMIT_AND_SHARE = 2
CALCRULE_ID_FRANCHISE_DEDUCTIBLE_AND_LIMIT = 3
CALCRULE_ID_DEDUCTIBLE_AND_LIMIT_PERCENT_TIV = 4
CALCRULE_ID_DEDUCTIBLE_AND_LIMIT_PERCENT_LOSS = 5
CALCRULE_ID_DEDUCTIBLE_PERCENT_TIV = 6
CALCRULE_ID_LIMIT_AND_MAX_DEDUCTIBLE = 7
CALCRULE_ID_LIMIT_AND_MIN_DEDUCTIBLE = 8
CALCRULE_ID_LIMIT_WITH_DEDUCTIBLE_PERCENT_LIMIT = 9
CALCRULE_ID_MAX_DEDUCTIBLE = 10
CALCRULE_ID_MIN_DEDUCTIBLE = 11
CALCRULE_ID_DEDUCTIBLE_ONLY = 12
CALCRULE_ID_MAIN_AND_MAX_DEDUCTIBLE = 13
CALCRULE_ID_LIMIT_ONLY = 14
CALCRULE_ID_LIMIT_PERCENT_LOSS = 15
CALCRULE_ID_DEDUCTIBLE_PERCENT_LOSS = 16
CALCRULE_ID_DEDUCTIBLE_PERCENT_LOSS_ATTACHMENT_LIMIT_AND_SHARE = 17
CALCRULE_ID_DEDUCTIBLE_PERCENT_TIV_ATTACHMENT_LIMIT_AND_SHARE = 18
CALCRULE_ID_DEDUCTIBLE_PERCENT_LOSS_WITH_MIN_AND_MAX = 19
CALCRULE_ID_REVERSE_FRANCHISE_DEDUCTIBLE = 20
CALCRULE_ID_SHARE_AND_LIMIT = 21
CALCRULE_ID_QUOTA_SHARE = 22
CALCRULE_ID_OCCURRENCE_LIMIT_AND_SHARE = 23
CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS = 24
CALCRULE_ID_FACULTATIVE_WITH_POLICY_SHARE = 25

NO_ALLOCATION_ALLOC_ID = 0
ALLOCATE_TO_ITEMS_BY_GUL_ALLOC_ID = 1
ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID = 2

BUILDING_COVERAGE_TYPE_ID = 1
OTHER_BUILDING_COVERAGE_TYPE_ID = 2
CONTENTS_COVERAGE_TYPE_ID = 3
TIME_COVERAGE_TYPE_ID = 4
COVERAGE_TYPES = [
    BUILDING_COVERAGE_TYPE_ID,
    OTHER_BUILDING_COVERAGE_TYPE_ID,
    CONTENTS_COVERAGE_TYPE_ID,
    TIME_COVERAGE_TYPE_ID]

PERIL_WIND = 1
PERILS = [PERIL_WIND]

GUL_INPUTS_FILES = [
    'coverages',
    'gulsummaryxref',
    'items']

IL_INPUTS_FILES = [
    'fm_policytc',
    'fm_profile',
    'fm_programme',
    'fm_xref',
    'fmsummaryxref']

OPTIONAL_INPUTS_FILES = [
    'events']

NOT_SET_ID = -1
LARGE_VALUE = 9999999999999


# --- OED constants --------------------------------------------------------- #
RI_INFO_REQUIRED_COLS = [
    'ReinsNumber',
    'PlacedPercent',
    'InuringPriority',
    'ReinsType',
    'RiskLevel',
    'ReinsCurrency'
]

RI_SCOPE_REQUIRED_COLS = {
    'ReinsNumber',
}

RI_INFO_DTYPES = {
    'ReinsNumber': "int",
    'CededPercent': "float",
    'RiskLimit': "float",
    'RiskAttachment': "float",
    'OccLimit': "float",
    'OccAttachment': "float",
    'InuringPriority': "int",
    'ReinsType': "str",
    'RiskLevel': "str",
    'PlacedPercent': "float",
    'TreatyShare': "float",
    'ReinsCurrency': 'str'
}

RI_SCOPE_DTYPES = {
    'ReinsNumber': "int",
    'PortNumber': "str",
    'AccNumber': "str",
    'LocGroup': "str",
    'PolNumber': "str",
    'LocNumber': "str",
    'CedantName': "str",
    'ProducerName': "str",
    'LOB': "str",
    'CountryCode': "str",
    'ReinsTag': "str",
    'CededPercent': "float"
}


POLICYITEM_LEVEL = 0
LOCATION_LEVEL = 1
POLICY_LEVEL = 2
ACCOUNT_LEVEL = 3
OCCURRENCE_LEVEL = 4

PASSTHROUGH_NODE_TYPE = 1
NOLOSS_NODE_TYPE = 1

REINS_TYPE_FAC = "FAC"
REINS_TYPE_QUOTA_SHARE = "QS"
REINS_TYPE_SURPLUS_SHARE = "SS"
REINS_TYPE_PER_RISK = "PR"
REINS_TYPE_CAT_XL = "CXL"
REINS_TYPE_AGG_XL = "AXL"
REINS_TYPES = [
    REINS_TYPE_FAC,
    REINS_TYPE_QUOTA_SHARE,
    REINS_TYPE_SURPLUS_SHARE,
    REINS_TYPE_PER_RISK,
    REINS_TYPE_CAT_XL
]

REINS_RISK_LEVEL_PORTFOLIO = "SEL"
REINS_RISK_LEVEL_LOCATION = "LOC"
REINS_RISK_LEVEL_LOCATION_GROUP = "LGR"
REINS_RISK_LEVEL_POLICY = "POL"
REINS_RISK_LEVEL_ACCOUNT = "ACC"
REINS_RISK_LEVELS = [
    REINS_RISK_LEVEL_LOCATION,
    REINS_RISK_LEVEL_LOCATION_GROUP,
    REINS_RISK_LEVEL_POLICY,
    REINS_RISK_LEVEL_ACCOUNT,
    REINS_RISK_LEVEL_PORTFOLIO,
]


SUPPORTED_RISK_LEVELS = {
    REINS_TYPE_FAC: [REINS_RISK_LEVEL_LOCATION, REINS_RISK_LEVEL_LOCATION_GROUP, REINS_RISK_LEVEL_POLICY, REINS_RISK_LEVEL_ACCOUNT],
    REINS_TYPE_SURPLUS_SHARE: [REINS_RISK_LEVEL_LOCATION, REINS_RISK_LEVEL_LOCATION_GROUP, REINS_RISK_LEVEL_POLICY, REINS_RISK_LEVEL_ACCOUNT],
    REINS_TYPE_PER_RISK: [REINS_RISK_LEVEL_LOCATION, REINS_RISK_LEVEL_LOCATION_GROUP, REINS_RISK_LEVEL_POLICY, REINS_RISK_LEVEL_ACCOUNT],
    REINS_TYPE_CAT_XL: [REINS_RISK_LEVEL_PORTFOLIO],
    REINS_TYPE_QUOTA_SHARE: REINS_RISK_LEVELS
}

# Subset of the fields that are currently used
OED_ACCOUNT_FIELDS = [
    'PortNumber',
    'AccNumber',
    'PolNumber',
    'AccPeril',
    'AccDed6All',
    'AccLimit6All'
]

OED_LOCATION_FIELDS = [
    'AccNumber',
    'LocGroup',
    'LocNumber',
    'LocDed6All',
    'LocLimit6All',
    'BuildingTIV',
    'OtherTIV',
    'ContentsTIV',
    'BITIV'
]

OED_REINS_INFO_FIELDS = [
    'ReinsNumber',
    'CededPercent',
    'RiskLimit',
    'RiskAttachment',
    'OccLimit',
    'OccAttachment',
    'InuringPriority',
    'ReinsType',
    'RiskLevel',
    'PlacedPercent',
    'TreatyShare',
    'ReinsCurrency',
]

OED_REINS_SCOPE_FIELDS = [
    'ReinsNumber',
    'PortNumber',
    'AccNumber',
    'PolNumber',
    'LocGroup',
    'LocNumber',
    'CedantName',
    'ProducerName',
    'LOB',
    'CountryCode',
    'ReinsTag',
    'CededPercent'
]



from contextlib import ExitStack
import logging
from pathlib import Path
import sys
from numba import njit
from numba.core import types
from numba.typed import Dict
import numpy as np

from oasis_data_manager.filestore.backends.base import BaseStorage
from oasislmf.pytools.common.data import lossfactors_headers, lossfactors_dtype
from oasislmf.pytools.common.event_stream import mv_read
from .common import PLAFACTORS_FILE


logger = logging.getLogger(__name__)


def get_post_loss_amplification_factors(storage: BaseStorage, secondary_factor, ignore_file_type=set()):
    """
    Get Post Loss Amplification (PLA) factors mapped to event ID-item ID pair.
    Returns empty dictionary if uniform factor to apply across all losses has
    been given.

    lossfactors.bin is binary file with layout:
        reserved header (4-byte int),
        event ID 1 (4-byte int), number of amplification IDs for event ID 1 (4-byte int),
        amplification ID 1 (4-byte int), loss factor for amplification ID 1 (4-byte float),
        ...
        amplification ID n (4-byte int), loss factor for amplification ID n (4-byte float),
        event ID 2 (4-byte int), number of amplification IDs for event ID 2 (4-byte int),
        ...
        event ID N (4-byte int), number of amplification IDs for event ID N (4-byte int),
        amplification ID 1 (4-byte int), loss factor for amplification ID 1 (4-byte float),
        ...
        amplification ID n (4-byte int), loss factor for amplification ID n (4-byte float)

    Args:
        storage: (BaseStorage) the storage connector for fetching the model data
        secondary_factor (float): secondary factor to apply to post loss
          amplification
        ignore_file_type: set(str) file extension to ignore when loading

    Returns:
        plafactors (dict): event ID-item ID pairs mapped to amplification IDs
    """
    input_files = set(storage.listdir())
    if PLAFACTORS_FILE in input_files and 'bin' not in ignore_file_type:
        plafactors = read_lossfactors(storage.root_dir, set(["csv"]), PLAFACTORS_FILE)
        for key, value in plafactors.items():
            plafactors[key] = max(
                1 + (value - 1) * secondary_factor, 0.0
            )
        return plafactors
    else:
        raise FileNotFoundError(f"lossfactors.bin file not found at {storage.get_storage_url('', encode_params=False)[1]}")


def read_lossfactors(run_dir="", ignore_file_type=set(), filename=PLAFACTORS_FILE, use_stdin=False):
    """Load the correlations from the lossfactors file.
    Args:
        run_dir (str): path to lossfactors.bin file
        ignore_file_type (Set[str]): file extension to ignore when loading.
        filename (str | os.PathLike): lossfactors file name
        use_stdin (bool): Use standard input for file data, ignores run_dir/filename. Defaults to False.
    Returns:
        plafactors (dict): event ID-item ID pairs mapped to amplification IDs
    """
    int32_itemsize = np.dtype(np.int32).itemsize
    float32_itemsize = np.dtype(np.float32).itemsize

    @njit(cache=True, error_model="numpy")
    def _read_bin(lossfactors, plafactors):
        cursor = 0
        opts, cursor = mv_read(lossfactors, cursor, np.int32, int32_itemsize)

        valid_buf = len(lossfactors)
        while cursor + (2 * int32_itemsize) <= valid_buf:
            event_id, cursor = mv_read(lossfactors, cursor, np.int32, int32_itemsize)
            count, cursor = mv_read(lossfactors, cursor, np.int32, int32_itemsize)
            for _ in range(count):
                if cursor + (int32_itemsize + float32_itemsize) > valid_buf:
                    break
                amplification_id, cursor = mv_read(lossfactors, cursor, np.int32, int32_itemsize)
                factor, cursor = mv_read(lossfactors, cursor, np.float32, float32_itemsize)
                plafactors[(event_id, amplification_id)] = factor

    @njit(cache=True, error_model="numpy")
    def _read_csv(lossfactors, plafactors):
        for row in lossfactors:
            plafactors[(row["event_id"], row["amplification_id"])] = row["factor"]

    def read_csv_lines(lines):
        # Check for header
        first_line_elements = [header.strip() for header in lines[0].strip().split(',')]
        has_header = first_line_elements == lossfactors_headers
        data_lines = lines[1:] if has_header else lines
        return np.loadtxt(
            data_lines,
            dtype=lossfactors_dtype,
            delimiter=",",
            ndmin=1
        )

    plafactors = Dict.empty(
        key_type=types.UniTuple(types.int64, 2), value_type=types.float64
    )

    supported_exts = ["bin", "csv"]

    if use_stdin:
        for ext in supported_exts:
            if ext in ignore_file_type:
                continue

            if ext == "bin":
                lossfactors = np.frombuffer(sys.stdin.buffer.read(), dtype=np.uint8)
                _read_bin(lossfactors, plafactors)
                return plafactors
            elif ext == "csv":
                lines = sys.stdin.readlines()
                lossfactors = read_csv_lines(lines)
                _read_csv(lossfactors, plafactors)
                return plafactors
            else:
                raise RuntimeError(f"Cannot read lossfactors file of type {ext}. Not Implemented.")
        raise RuntimeError(
            f'lossfactors data not readable with use_stdin={use_stdin}, run_dir={run_dir}, filename={filename}. Ignoring files with ext {ignore_file_type}.')

    for ext in supported_exts:
        if ext in ignore_file_type:
            continue

        lossfactors_file = Path(run_dir, filename).with_suffix("." + ext)
        if not lossfactors_file.exists():
            continue

        if ext == "bin":
            lossfactors = np.memmap(lossfactors_file, dtype=np.uint8, mode='r')
            _read_bin(lossfactors, plafactors)
            return plafactors
        elif ext == "csv":
            with ExitStack() as stack:
                fin = stack.enter_context(open(lossfactors_file, "r"))
                lines = fin.readlines()
            lossfactors = read_csv_lines(lines)
            _read_csv(lossfactors, plafactors)
            return plafactors
        else:
            raise RuntimeError(f"Cannot read lossfactors file of type {ext}. Not Implemented.")

    raise RuntimeError(
        f'lossfactors data not readable with use_stdin={use_stdin}, run_dir={run_dir}, filename={filename}. Ignoring files with ext {ignore_file_type}.')



# aal/manager.py

import logging
import numpy as np
import numba as nb
import os
from contextlib import ExitStack
from pathlib import Path
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from oasislmf.pytools.aal.data import AAL_meanonly_dtype, AAL_meanonly_fmt, AAL_meanonly_headers, AAL_dtype, AAL_fmt, AAL_headers, ALCT_dtype, ALCT_fmt, ALCT_headers
from oasislmf.pytools.common.data import (DEFAULT_BUFFER_SIZE, MEAN_TYPE_ANALYTICAL, MEAN_TYPE_SAMPLE, oasis_int, oasis_float,
                                          oasis_int_size, oasis_float_size, write_ndarray_to_fmt_csv)
from oasislmf.pytools.common.event_stream import (MEAN_IDX, MAX_LOSS_IDX, NUMBER_OF_AFFECTED_RISK_IDX, SUMMARY_STREAM_ID,
                                                  init_streams_in, mv_read)
from oasislmf.pytools.common.input_files import read_occurrence, read_periods
from oasislmf.pytools.common.utils.nb_heapq import heap_pop, heap_push, init_heap
from oasislmf.pytools.utils import redirect_logging


logger = logging.getLogger(__name__)

# Total amount of memory AAL summary index should use before raising an error (GB)
OASIS_AAL_MEMORY = float(os.environ["OASIS_AAL_MEMORY"]) if "OASIS_AAL_MEMORY" in os.environ else 4

# Similar to aal_rec in ktools
# summary_id can be infered from index
# type can be infered from which array is using it
_AAL_REC_DTYPE = np.dtype([
    ('mean', np.float64),
    ('mean_squared', np.float64),
])

# Similar to aal_rec_period
_AAL_REC_PERIOD_DTYPE = np.dtype(
    _AAL_REC_DTYPE.descr + [('mean_period', np.float64)]
)

_SUMMARIES_DTYPE = np.dtype([
    ("summary_id", np.int32),
    ("file_idx", np.int32),
    ("period_no", np.int32),
    ("file_offset", np.int64),
])
_SUMMARIES_DTYPE_size = _SUMMARIES_DTYPE.itemsize


@nb.njit(cache=True, error_model="numpy")
def process_bin_file(
    fbin,
    offset,
    occ_map,
    summaries_data,
    summaries_idx,
    file_index,
):
    """Reads summary<n>.bin file event_ids and summary_ids to populate summaries_data
    Args:
        fbin (np.memmap): summary binary memmap
        offset (int): file offset to read from
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
        summaries_data (ndarray[_SUMMARIES_DTYPE]): Index summary data (summaries.idx data)
        summaries_idx (int): current index reached in summaries_data
        file_index (int): Summary bin file index
    Returns:
        summaries_idx (int): current index reached in summaries_data
        resize_flag (bool): flag to indicate whether to resize summaries_data when full
        offset (int): file offset to read from
    """
    while offset < len(fbin):
        cursor = offset
        event_id, cursor = mv_read(fbin, cursor, oasis_int, oasis_int_size)
        summary_id, cursor = mv_read(fbin, cursor, oasis_int, oasis_int_size)

        if event_id not in occ_map:
            # If the event_id doesn't exist in occ_map, continue with the next
            offset = cursor
            # Skip over Expval and losses
            _, offset = mv_read(fbin, offset, oasis_float, oasis_float_size)
            offset = skip_losses(fbin, offset)
            continue

        # Get the number of rows for the current event_id
        n_rows = occ_map[event_id].shape[0]

        if summaries_idx + n_rows >= len(summaries_data):
            # Resize array if full
            return summaries_idx, True, offset

        # Now fill the summaries_data with the rows that match the current event_id
        current_row = 0
        for row in occ_map[event_id]:
            summaries_data[summaries_idx]["summary_id"] = summary_id
            summaries_data[summaries_idx]["file_idx"] = file_index
            summaries_data[summaries_idx]["period_no"] = row["period_no"]
            summaries_data[summaries_idx]["file_offset"] = offset
            summaries_idx += 1
            current_row += 1
            if current_row >= n_rows:
                break

        offset = cursor
        # Read Expval
        _, offset = mv_read(fbin, offset, oasis_float, oasis_float_size)
        # Skip over losses
        offset = skip_losses(fbin, offset)

    return summaries_idx, False, offset


def sort_and_save_chunk(summaries_data, temp_file_path):
    """Sort a chunk of summaries data and save it to a temporary file.
    Args:
        summaries_data (ndarray[_SUMMARIES_DTYPE]): Indexed summary data
        temp_file_path (str | os.PathLike): Path to temporary file
    """
    sort_columns = ["file_idx", "period_no", "summary_id"]
    sorted_indices = np.lexsort([summaries_data[col] for col in sort_columns])
    sorted_chunk = summaries_data[sorted_indices]
    sorted_chunk.tofile(temp_file_path)


@nb.njit(cache=True, error_model="numpy")
def merge_sorted_chunks(memmaps):
    """
    Merge sorted chunks using a k-way merge algorithm and yield next smallest row
    Args:
        memmaps (List[np.memmap]): List of temporary file memmaps
    Yields:
        smallest_row (ndarray[_SUMMARIES_DTYPE]): yields the next smallest row from sorted summaries partial files
    """
    min_heap = init_heap(num_compare=3)
    size = 0
    # Initialize the min_heap with the first row of each memmap
    for i, mmap in enumerate(memmaps):
        if len(mmap) > 0:
            first_row = mmap[0]
            min_heap, size = heap_push(min_heap, size, np.array(
                [first_row["summary_id"], first_row["period_no"], first_row["file_idx"], i, 0],
                dtype=np.int32
            ))

    # Perform the k-way merge
    while size > 0:
        # The min heap will store the smallest row at the top when popped
        element, min_heap, size = heap_pop(min_heap, size)
        file_idx = element[3]
        row_num = element[4]
        smallest_row = memmaps[file_idx][row_num]
        yield smallest_row

        # Push the next row from the same file into the heap if there are any more rows
        if row_num + 1 < len(memmaps[file_idx]):
            next_row = memmaps[file_idx][row_num + 1]
            min_heap, size = heap_push(min_heap, size, np.array(
                [next_row["summary_id"], next_row["period_no"], next_row["file_idx"], file_idx, row_num + 1],
                dtype=np.int32
            ))


def get_summaries_data(
    path,
    files_handles,
    occ_map,
    aal_max_memory
):
    """Gets the indexed summaries data, ordered with k-way merge if not enough memory
    Args:
        path (os.PathLike): Path to the workspace folder containing summary binaries
        files_handles (List[np.memmap]): List of memmaps for summary files data
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
        aal_max_memory (float): OASIS_AAL_MEMORY value (has to be passed in as numba won't update from environment variable)
    Returns:
        memmaps (List[np.memmap]): List of temporary file memmaps
        max_summary_id (int): Max summary ID
    """
    # Remove existing temp bdat files if exists
    for temp_file in path.glob("indexed_summaries.part*.bdat"):
        os.remove(temp_file)

    buffer_size = int(((aal_max_memory * (1024**3) // _SUMMARIES_DTYPE_size)))
    temp_files = []
    chunk_index = 0

    summaries_data = np.empty(buffer_size, dtype=_SUMMARIES_DTYPE)
    summaries_idx = 0
    max_summary_id = 0
    for file_index, fbin in enumerate(files_handles):
        offset = oasis_int_size * 3  # Summary stream header size

        while True:
            summaries_idx, resize_flag, offset = process_bin_file(
                fbin,
                offset,
                occ_map,
                summaries_data,
                summaries_idx,
                file_index,
            )

            # Write new summaries partial file when buffer size or end of summary file reached
            if resize_flag:
                temp_file_path = Path(path, f"indexed_summaries.part{chunk_index}.bdat")
                summaries_data = summaries_data[:summaries_idx]
                sort_and_save_chunk(summaries_data, temp_file_path)
                temp_files.append(temp_file_path)
                chunk_index += 1
                summaries_idx = 0
                max_summary_id = max(max_summary_id, np.max(summaries_data["summary_id"]))

            # End of file, move to next file
            if offset >= len(fbin):
                break

    # Write remaining summaries data to temporary file
    temp_file_path = Path(path, f"indexed_summaries.part{chunk_index}.bdat")
    summaries_data = summaries_data[:summaries_idx]
    sort_and_save_chunk(summaries_data, temp_file_path)
    max_summary_id = max(max_summary_id, np.max(summaries_data["summary_id"]))
    temp_files.append(temp_file_path)

    memmaps = [np.memmap(temp_file, mode="r", dtype=_SUMMARIES_DTYPE) for temp_file in temp_files]

    return memmaps, max_summary_id


def summary_index(path, occ_map, stack):
    """Index the summary binary outputs
    Args:
        path (os.PathLike): Path to the workspace folder containing summary binaries
        occ_map (nb.typed.Dict): numpy map of event_id, period_no, occ_date_id from the occurrence file
        stack (ExitStack): Exit stack
    Returns:
        files_handles (List[np.memmap]): List of memmaps for summary files data
        sample_size (int): Sample size
        max_summary_id (int): Max summary ID
        memmaps (List[np.memmap]): List of temporary file memmaps
    """
    # work folder for aal files
    aal_files_folder = Path(path, "aal_files")
    aal_files_folder.mkdir(parents=False, exist_ok=True)

    # Find summary binary files
    files = [file for file in path.glob("*.bin")]
    files_handles = [np.memmap(file, mode="r", dtype="u1") for file in files]

    streams_in, (stream_source_type, stream_agg_type, sample_size) = init_streams_in(files, stack)
    if stream_source_type != SUMMARY_STREAM_ID:
        raise RuntimeError(f"Error: Not a summary stream type {stream_source_type}")

    memmaps, max_summary_id = get_summaries_data(
        aal_files_folder,
        files_handles,
        occ_map,
        OASIS_AAL_MEMORY
    )
    return files_handles, sample_size, max_summary_id, memmaps


def read_input_files(run_dir):
    """Reads all input files and returns a dict of relevant data
    Args:
        run_dir (str | os.PathLike): Path to directory containing required files structure
    Returns:
        file_data (Dict[str, Any]): A dict of relevent data extracted from files
    """
    occ_map, date_algorithm, granular_date, no_of_periods = read_occurrence(Path(run_dir, "input"))
    period_weights = read_periods(no_of_periods, Path(run_dir, "input"))

    file_data = {
        "occ_map": occ_map,
        "date_algorithm": date_algorithm,
        "granular_date": granular_date,
        "no_of_periods": no_of_periods,
        "period_weights": period_weights,
    }
    return file_data


def get_num_subsets(alct, sample_size, max_summary_id):
    """Gets the number of subsets required to generates the Sample AAL np map for subset sizes up to sample_size
    Example: sample_size[10], max_summary_id[2] generates following ndarray
    [
        #   subset_size, mean,  mean_squared, mean_period
        [0, 0, 0],  # subset_size = 1 , summary_id = 1
        [0, 0, 0],  # subset_size = 1 , summary_id = 2
        [0, 0, 0],  # subset_size = 2 , summary_id = 1
        [0, 0, 0],  # subset_size = 2 , summary_id = 2
        [0, 0, 0],  # subset_size = 4 , summary_id = 1
        [0, 0, 0],  # subset_size = 4 , summary_id = 2
        [0, 0, 0],  # subset_size = 10 , summary_id = 1, subset_size = sample_size
        [0, 0, 0],  # subset_size = 10 , summary_id = 2, subset_size = sample_size
    ]
    Subset_size is implicit based on position in array, grouped by max_summary_id
    So first two arrays are subset_size 2^0 = 1
    The next two arrays are subset_size 2^1 = 2
    The next two arrays are subset_size 2^2 = 4
    The last two arrays are subset_size = sample_size = 10
    Doesn't generate one with subset_size 8 as double that is larger than sample_size
    Therefore this function returns 4, and the sample aal array is 4 * 2
    Args:
        alct (bool): Boolean for ALCT output
        sample_size (int): Sample size
        max_summary_id (int): Max summary ID
    Returns:
        num_subsets (int): Number of subsets
    """
    i = 0
    if alct and sample_size > 1:
        while ((1 << i) + ((1 << i) - 1)) <= sample_size:
            i += 1
    return i + 1


@nb.njit(cache=True, fastmath=True, error_model="numpy")
def get_weighted_means(
    vec_sample_sum_loss,
    weighting,
    sidx,
    end_sidx,
):
    """Get sum of weighted mean and weighted mean_squared
    Args:
        vec_sample_sum_loss (ndarray[_AAL_REC_DTYPE]): Vector for sample sum losses
        weighting (float): Weighting value
        sidx (int): start index
        end_sidx (int): end index
    Returns:
        weighted_mean (float): Sum weighted mean
        weighted_mean_squared (float): Sum weighted mean squared
    """
    weighted_mean = 0
    weighted_mean_squared = 0
    while sidx < end_sidx:
        sumloss = vec_sample_sum_loss[sidx]
        weighted_mean += sumloss * weighting
        weighted_mean_squared += sumloss * sumloss * weighting
        sidx += 1
    return weighted_mean, weighted_mean_squared


@nb.njit(cache=True, error_model="numpy")
def do_calc_end(
    period_no,
    no_of_periods,
    period_weights,
    sample_size,
    curr_summary_id,
    max_summary_id,
    vec_analytical_aal,
    vecs_sample_aal,
    vec_used_summary_id,
    vec_sample_sum_loss,
):
    """Updates Analytical and Sample AAL vectors from sample sum losses
    Args:
        period_no (int): Period Number
        no_of_periods (int): Number of periods
        period_weights (ndarray[periods_dtype]): Period Weights
        sample_size (int): Sample Size
        curr_summary_id (int): Current summary_id
        max_summary_id (int): Max summary_id
        vec_analytical_aal (ndarray[_AAL_REC_DTYPE]): Vector for Analytical AAL
        vecs_sample_aal (ndarray[_AAL_REC_PERIODS_DTYPE]): Vector for Sample AAL
        vec_used_summary_id (ndarray[bool]): vector to store if summary_id is used
        vec_sample_sum_loss (ndarray[_AAL_REC_DTYPE]): Vector for sample sum losses
    """
    # Get weighting
    weighting = 1
    if no_of_periods > 0:
        # period_no in period_weights
        if period_no > 0 and period_no <= no_of_periods:
            weighting = period_weights[period_no - 1][1] * no_of_periods
        else:
            weighting = 0

    # Update Analytical AAL
    mean = vec_sample_sum_loss[0]  # 0 index is where the analytical mean is stored
    vec_used_summary_id[curr_summary_id - 1] = True

    vec_analytical_aal[curr_summary_id - 1]["mean"] += mean * weighting
    vec_analytical_aal[curr_summary_id - 1]["mean_squared"] += mean * mean * weighting

    # Update Sample AAL
    # Get relevant indexes for curr_summary_id
    len_sample_aal = len(vecs_sample_aal)
    num_subsets = len_sample_aal // max_summary_id
    idxs = [i * max_summary_id + (curr_summary_id - 1) for i in range(num_subsets)]

    # Get sample aal idx for sample_size
    last_sample_aal = vecs_sample_aal[idxs[-1]]

    total_mean_by_period = 0
    sidx = 1
    aal_idx = 0
    while sidx < sample_size + 1:
        # Iterate through aal_idx except the last one which is subset_size == sample_size
        while aal_idx < num_subsets - 1:
            curr_sample_aal = vecs_sample_aal[idxs[aal_idx]]

            # Calculate the subset_size and assign to sidx
            sidx = 1 << aal_idx
            end_sidx = sidx << 1

            # Traverse sidx == subset_size to sidx == subset_size * 2
            weighted_mean, weighted_mean_squared = get_weighted_means(
                vec_sample_sum_loss,
                weighting,
                sidx,
                end_sidx
            )

            # Update sample size Sample AAL
            last_sample_aal["mean"] += weighted_mean
            last_sample_aal["mean_squared"] += weighted_mean_squared
            total_mean_by_period += weighted_mean

            # Update current Sample AAL
            curr_sample_aal["mean"] += weighted_mean
            curr_sample_aal["mean_squared"] += weighted_mean_squared
            # Update current Sample AAL mean_period
            curr_sample_aal["mean_period"] += weighted_mean * weighted_mean

            sidx = end_sidx
            aal_idx += 1
        # Update sample size Sample AAL
        mean = vec_sample_sum_loss[sidx]
        total_mean_by_period += mean * weighting
        last_sample_aal["mean"] += mean * weighting
        last_sample_aal["mean_squared"] += mean * mean * weighting
        sidx += 1
    # Update sample size Sample AAL mean_period
    last_sample_aal["mean_period"] += total_mean_by_period * total_mean_by_period
    vec_sample_sum_loss.fill(0)


@nb.njit(cache=True, error_model="numpy")
def read_losses(summary_fin, cursor, vec_sample_sum_loss):
    """Read losses from summary_fin starting at cursor, populate vec_sample_sum_loss
    Args:
        summary_fin (np.memmap): summary file memmap
        cursor (int): data offset for reading binary files
        (ndarray[_AAL_REC_DTYPE]): Vector for sample sum losses
    Returns:
        cursor (int): data offset for reading binary files
    """
    # Max losses is sample_size + num special sidxs
    valid_buff = len(summary_fin)
    while True:
        if valid_buff - cursor < oasis_int_size + oasis_float_size:
            raise RuntimeError("Error: broken summary file, not enough data")
        sidx, cursor = mv_read(summary_fin, cursor, oasis_int, oasis_int_size)
        loss, cursor = mv_read(summary_fin, cursor, oasis_float, oasis_float_size)

        if sidx == 0:
            break
        if sidx == NUMBER_OF_AFFECTED_RISK_IDX or sidx == MAX_LOSS_IDX:
            continue
        if sidx == MEAN_IDX:
            sidx = 0
        vec_sample_sum_loss[sidx] += loss
    return cursor


@nb.njit(cache=True, error_model="numpy")
def skip_losses(summary_fin, cursor):
    """Skip through losses in summary_fin starting at cursor
    Args:
        summary_fin (np.memmap): summary file memmap
        cursor (int): data offset for reading binary files
    Returns:
        cursor (int): data offset for reading binary files
    """
    valid_buff = len(summary_fin)
    sidx = 1
    while sidx:
        if valid_buff - cursor < oasis_int_size + oasis_float_size:
            raise RuntimeError("Error: broken summary file, not enough data")
        sidx, cursor = mv_read(summary_fin, cursor, oasis_int, oasis_int_size)
        cursor += oasis_float_size
    return cursor


@nb.njit(cache=True, error_model="numpy")
def run_aal(
    memmaps,
    no_of_periods,
    period_weights,
    sample_size,
    max_summary_id,
    files_handles,
    vec_analytical_aal,
    vecs_sample_aal,
    vec_used_summary_id,
):
    """Run AAL calculation loop to populate vec data
    Args:
        memmaps (List[np.memmap]): List of temporary file memmaps
        no_of_periods (int): Number of periods
        period_weights (ndarray[periods_dtype]): Period Weights
        sample_size (int): Sample Size
        max_summary_id (int): Max summary_id
        files_handles (List[np.memmap]): List of memmaps for summary files data
        vec_analytical_aal (ndarray[_AAL_REC_DTYPE]): Vector for Analytical AAL
        vecs_sample_aal (ndarray[_AAL_REC_PERIODS_DTYPE]): Vector for Sample AAL
        vec_used_summary_id (ndarray[bool]): vector to store if summary_id is used
    """
    if len(memmaps) == 0:
        raise ValueError("File is empty or missing data")

    # Index 0 is mean
    vec_sample_sum_loss = np.zeros(sample_size + 1, dtype=np.float64)
    last_summary_id = -1
    last_period_no = -1

    for line in merge_sorted_chunks(memmaps):
        summary_id = line["summary_id"]
        file_idx = line["file_idx"]
        period_no = line["period_no"]
        file_offset = line["file_offset"]

        if last_summary_id != summary_id:
            if last_summary_id != -1:
                do_calc_end(
                    last_period_no,
                    no_of_periods,
                    period_weights,
                    sample_size,
                    last_summary_id,
                    max_summary_id,
                    vec_analytical_aal,
                    vecs_sample_aal,
                    vec_used_summary_id,
                    vec_sample_sum_loss,
                )
            last_period_no = period_no
            last_summary_id = summary_id
        elif last_period_no != period_no:
            if last_period_no != -1:
                do_calc_end(
                    last_period_no,
                    no_of_periods,
                    period_weights,
                    sample_size,
                    last_summary_id,
                    max_summary_id,
                    vec_analytical_aal,
                    vecs_sample_aal,
                    vec_used_summary_id,
                    vec_sample_sum_loss,
                )
            last_period_no = period_no
        summary_fin = files_handles[file_idx]

        # Read summary header values (event_id, summary_id, expval)
        cursor = file_offset + (2 * oasis_int_size) + oasis_float_size

        read_losses(summary_fin, cursor, vec_sample_sum_loss)

    if last_summary_id != -1:
        do_calc_end(
            last_period_no,
            no_of_periods,
            period_weights,
            sample_size,
            last_summary_id,
            max_summary_id,
            vec_analytical_aal,
            vecs_sample_aal,
            vec_used_summary_id,
            vec_sample_sum_loss,
        )


@nb.njit(cache=True, fastmath=True, error_model="numpy")
def calculate_mean_stddev(
    observable_sum,
    observable_squared_sum,
    number_of_observations
):
    """Compute the mean and standard deviation from the sum and squared sum of an observable
    Args:
        observable_sum (ndarray[oasis_float]): Observable sum
        observable_squared_sum (ndarray[oasis_float]): Observable squared sum
        number_of_observations (int | ndarray[int]): number of observations
    Returns:
        mean (ndarray[oasis_float]): Mean
        std (ndarray[oasis_float]): Standard Deviation
    """
    mean = observable_sum / number_of_observations
    std = np.sqrt(
        (
            observable_squared_sum - (observable_sum * observable_sum)
            / number_of_observations
        ) / (number_of_observations - 1)
    )
    return mean, std


@nb.njit(cache=True, error_model="numpy")
def get_aal_data(
    vec_analytical_aal,
    vecs_sample_aal,
    vec_used_summary_id,
    sample_size,
    no_of_periods
):
    """Generate AAL csv data
    Args:
        vec_analytical_aal (ndarray[_AAL_REC_DTYPE]): Vector for Analytical AAL
        vecs_sample_aal (ndarray[_AAL_REC_PERIODS_DTYPE]): Vector for Sample AAL
        vec_used_summary_id (ndarray[bool]): vector to store if summary_id is used
        sample_size (int): Sample Size
        no_of_periods (int): Number of periods
    Returns:
        aal_data (List[Tuple]): AAL csv data
    """
    aal_data = []
    assert len(vec_analytical_aal) == len(vecs_sample_aal), \
        f"Lengths of analytical ({len(vec_analytical_aal)}) and sample ({len(vecs_sample_aal)}) aal data differ"
    mean_analytical, std_analytical = calculate_mean_stddev(
        vec_analytical_aal["mean"],
        vec_analytical_aal["mean_squared"],
        no_of_periods,
    )

    mean_sample, std_sample = calculate_mean_stddev(
        vecs_sample_aal["mean"],
        vecs_sample_aal["mean_squared"],
        no_of_periods * sample_size,
    )

    for summary_idx in range(len(vec_analytical_aal)):
        if not vec_used_summary_id[summary_idx]:
            continue
        aal_data.append((summary_idx + 1, MEAN_TYPE_ANALYTICAL, mean_analytical[summary_idx], std_analytical[summary_idx]))
    for summary_idx in range(len(vecs_sample_aal)):
        if not vec_used_summary_id[summary_idx]:
            continue
        aal_data.append((summary_idx + 1, MEAN_TYPE_SAMPLE, mean_sample[summary_idx], std_sample[summary_idx]))

    return aal_data


@nb.njit(cache=True, error_model="numpy")
def get_aal_data_meanonly(
    vec_analytical_aal,
    vecs_sample_aal,
    vec_used_summary_id,
    sample_size,
    no_of_periods
):
    """Generate AAL csv data
    Args:
        vec_analytical_aal (ndarray[_AAL_REC_DTYPE]): Vector for Analytical AAL
        vecs_sample_aal (ndarray[_AAL_REC_PERIODS_DTYPE]): Vector for Sample AAL
        vec_used_summary_id (ndarray[bool]): vector to store if summary_id is used
        sample_size (int): Sample Size
        no_of_periods (int): Number of periods
    Returns:
        aal_data (List[Tuple]): AAL csv data
    """
    aal_data = []
    assert len(vec_analytical_aal) == len(vecs_sample_aal), \
        f"Lengths of analytical ({len(vec_analytical_aal)}) and sample ({len(vecs_sample_aal)}) aal data differ"
    mean_analytical, std_analytical = calculate_mean_stddev(
        vec_analytical_aal["mean"],
        vec_analytical_aal["mean_squared"],
        no_of_periods,
    )

    mean_sample, std_sample = calculate_mean_stddev(
        vecs_sample_aal["mean"],
        vecs_sample_aal["mean_squared"],
        no_of_periods * sample_size,
    )

    # aalmeanonlycalc orders output data differently, this if condition is here to match the output to the ktools output
    for summary_idx in range(len(vec_analytical_aal)):
        if not vec_used_summary_id[summary_idx]:
            continue
        aal_data.append((summary_idx + 1, MEAN_TYPE_ANALYTICAL, mean_analytical[summary_idx]))
        aal_data.append((summary_idx + 1, MEAN_TYPE_SAMPLE, mean_sample[summary_idx]))

    return aal_data


@nb.njit(cache=True, fastmath=True, error_model="numpy")
def calculate_confidence_interval(std_err, confidence_level):
    """Calculate the confidence interval based on standard error and confidence level.
    Args:
        std_err (float): The standard error.
        confidence_level (float): The confidence level (e.g., 0.95 for 95%).
    Returns:
        confidence interval (float): The confidence interval.
    """
    # Compute p-value above 0.5
    p_value = (1 + confidence_level) / 2
    p_value = np.sqrt(-2 * np.log(1 - p_value))

    # Approximation formula for z-value from Abramowitz & Stegun, Handbook
    # of Mathematical Functions: with Formulas, Graphs, and Mathematical
    # Tables, Dover Publications (1965), eq. 26.2.23
    # Also see John D. Cook Consulting, https://www.johndcook.com/blog/cpp_phi_inverse/
    c = np.array([2.515517, 0.802853, 0.010328])
    d = np.array([1.432788, 0.189269, 0.001308])
    z_value = p_value - (
        ((c[2] * p_value + c[1]) * p_value + c[0]) /
        (((d[2] * p_value + d[1]) * p_value + d[0]) * p_value + 1)
    )
    # Return the confidence interval
    return std_err * z_value


@nb.njit(cache=True, error_model="numpy")
def get_alct_data(
    vecs_sample_aal,
    max_summary_id,
    sample_size,
    no_of_periods,
    confidence,
):
    """Generate ALCT csv data
    Args:
        vecs_sample_aal (ndarray[_AAL_REC_PERIODS_DTYPE]): Vector for Sample AAL
        max_summary_id (int): Max summary_id
        sample_size (int): Sample Size
        no_of_periods (int): Number of periods
        confidence (float): Confidence level between 0 and 1, default 0.95
    Returns:
        alct_data (List[List]): ALCT csv data
    """
    alct_data = []

    num_subsets = len(vecs_sample_aal) // max_summary_id
    # Generate the subset sizes (last one is always sample_size)
    subset_sizes = np.array([2 ** i for i in range(num_subsets)])
    subset_sizes[-1] = sample_size

    for summary_id in range(1, max_summary_id + 1):
        # Get idxs for summary_id across all subset_sizes
        idxs = np.array([i * max_summary_id + (summary_id - 1) for i in range(num_subsets)])
        v_curr = vecs_sample_aal[idxs]

        mean, std = calculate_mean_stddev(
            v_curr["mean"],
            v_curr["mean_squared"],
            subset_sizes * no_of_periods,
        )
        mean_period = v_curr["mean_period"] / (subset_sizes * subset_sizes)

        var_vuln = (
            (v_curr["mean_squared"] - subset_sizes * mean_period)
            / (subset_sizes * no_of_periods - subset_sizes)
        ) / (subset_sizes * no_of_periods)
        var_haz = (
            subset_sizes * (mean_period - no_of_periods * mean * mean)
            / (no_of_periods - 1)
        ) / (subset_sizes * no_of_periods)

        std_err = np.sqrt(var_vuln)
        ci = calculate_confidence_interval(std_err, confidence)

        std_err_haz = np.sqrt(var_haz)
        std_err_vuln = np.sqrt(var_vuln)

        lower_ci = np.where(ci > 0, mean - ci, 0)
        upper_ci = np.where(ci > 0, mean + ci, 0)

        curr_data = np.column_stack((
            np.array([summary_id] * num_subsets),
            mean,
            std,
            subset_sizes,
            lower_ci,
            upper_ci,
            std_err, std_err / mean,
            var_haz, std_err_haz, std_err_haz / mean,
            var_vuln, std_err_vuln, std_err_vuln / mean,
        ))
        for row in curr_data:
            alct_data.append(row)
    return alct_data


def run(
    run_dir,
    subfolder,
    aal_output_file=None,
    alct_output_file=None,
    meanonly=False,
    noheader=False,
    confidence=0.95,
    output_format="csv",
):
    """Runs AAL calculations
    Args:
        run_dir (str | os.PathLike): Path to directory containing required files structure
        subfolder (str): Workspace subfolder inside <run_dir>/work/<subfolder>
        aal_output_file (str, optional): Path to AAL output file. Defaults to None
        alct_output_file (str, optional): Path to ALCT output file. Defaults to None
        meanonly (bool): Boolean value to output AAL with mean only
        noheader (bool): Boolean value to skip header in output file
        confidence (float): Confidence level between 0 and 1, default 0.95
        output_format (str): Output format extension. Defaults to "csv".
    """
    outmap = {
        "aal": {
            "compute": aal_output_file is not None,
            "file_path": aal_output_file,
            "fmt": AAL_fmt if not meanonly else AAL_meanonly_fmt,
            "headers": AAL_headers if not meanonly else AAL_meanonly_headers,
            "file": None,
            "dtype": AAL_dtype if not meanonly else AAL_meanonly_dtype,
        },
        "alct": {
            "compute": alct_output_file is not None,
            "file_path": alct_output_file,
            "fmt": ALCT_fmt,
            "headers": ALCT_headers,
            "file": None,
            "dtype": ALCT_dtype,
        },
    }

    output_format = "." + output_format
    output_binary = output_format == ".bin"
    output_parquet = output_format == ".parquet"
    # Check for correct suffix
    for path in [v["file_path"] for v in outmap.values()]:
        if path is None:
            continue
        if Path(path).suffix == "":  # Ignore suffix for pipes
            continue
        if (Path(path).suffix != output_format):
            raise ValueError(f"Invalid file extension for {output_format}, got {path},")

    if not all([v["compute"] for v in outmap.values()]):
        logger.warning("No output files specified")

    with ExitStack() as stack:
        workspace_folder = Path(run_dir, "work", subfolder)
        if not workspace_folder.is_dir():
            raise RuntimeError(f"Error: Unable to open directory {workspace_folder}")

        file_data = read_input_files(run_dir)
        files_handles, sample_size, max_summary_id, memmaps = summary_index(
            workspace_folder,
            file_data["occ_map"],
            stack
        )
        # aal vec are Indexed on summary_id - 1
        num_subsets = get_num_subsets(outmap["alct"]["compute"], sample_size, max_summary_id)
        vecs_sample_aal = np.zeros(num_subsets * max_summary_id, dtype=_AAL_REC_PERIOD_DTYPE)
        vec_analytical_aal = np.zeros(max_summary_id, dtype=_AAL_REC_DTYPE)
        vec_used_summary_id = np.zeros(max_summary_id, dtype=np.bool_)

        # Run AAL calculations, populate above vecs
        run_aal(
            memmaps,
            file_data["no_of_periods"],
            file_data["period_weights"],
            sample_size,
            max_summary_id,
            files_handles,
            vec_analytical_aal,  # unique in output_aal
            vecs_sample_aal,
            vec_used_summary_id,
        )

        # Initialise output files AAL
        if output_binary:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                out_file = stack.enter_context(open(outmap[out_type]["file_path"], 'wb'))
                outmap[out_type]["file"] = out_file
        elif output_parquet:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                temp_out_data = np.zeros(DEFAULT_BUFFER_SIZE, dtype=outmap[out_type]["dtype"])
                temp_df = pd.DataFrame(temp_out_data, columns=outmap[out_type]["headers"])
                temp_table = pa.Table.from_pandas(temp_df)
                out_file = pq.ParquetWriter(outmap[out_type]["file_path"], temp_table.schema)
                outmap[out_type]["file"] = out_file
        else:
            for out_type in outmap:
                if not outmap[out_type]["compute"]:
                    continue
                out_file = stack.enter_context(open(outmap[out_type]["file_path"], 'w'))
                if not noheader:
                    csv_headers = ','.join(outmap[out_type]["headers"])
                    out_file.write(csv_headers + '\n')
                outmap[out_type]["file"] = out_file

        def write_output(data, out_type):
            if output_binary:
                data.tofile(outmap[out_type]["file"])
            elif output_parquet:
                data_df = pd.DataFrame(data)
                data_table = pa.Table.from_pandas(data_df)
                outmap[out_type]["file"].write_table(data_table)
            else:
                write_ndarray_to_fmt_csv(
                    outmap[out_type]["file"],
                    data,
                    outmap[out_type]["headers"],
                    outmap[out_type]["fmt"]
                )

        if outmap["aal"]["compute"]:
            # Get Sample AAL data for subset_size == sample_size (last group of arrays)
            start_idx = (num_subsets - 1) * max_summary_id
            aal_data_func = get_aal_data_meanonly if meanonly else get_aal_data
            aal_data = aal_data_func(
                vec_analytical_aal,
                vecs_sample_aal[start_idx:],
                vec_used_summary_id,
                sample_size,
                file_data["no_of_periods"],
            )
            aal_data = np.array(aal_data, dtype=outmap["aal"]["dtype"])

            write_output(aal_data, "aal")
        if outmap["alct"]["compute"]:
            alct_data = get_alct_data(
                vecs_sample_aal,
                max_summary_id,
                sample_size,
                file_data["no_of_periods"],
                confidence
            )
            alct_data = np.array([tuple(arr) for arr in alct_data], dtype=outmap["alct"]["dtype"])

            write_output(alct_data, "alct")


@redirect_logging(exec_name='aalpy')
def main(run_dir='.', subfolder=None, aal=None, alct=None, meanonly=False, noheader=False, confidence=0.95, ext="csv", **kwargs):
    run(
        run_dir,
        subfolder,
        aal_output_file=aal,
        meanonly=meanonly,
        alct_output_file=alct,
        noheader=noheader,
        confidence=confidence,
        output_format=ext,
    )



__all__ = [
    'COVERAGE_TYPES',
    'SUPPORTED_COVERAGE_TYPES'
]

from collections import OrderedDict


# Property
COVT_BLD = 1
COVT_OTH = 2
COVT_CON = 3
COVT_BIT = 4
COVT_PDM = 5
COVT_ALL = 6
# Cyber
COVT_BINP = 7
COVT_CBI = 8
COVT_DIAS = 9
COVT_EXT = 10
COVT_FIN = 11
COVT_INRE = 12
COVT_LIAB = 13
COVT_REG = 14
COVT_ENO = 15
COVT_CYB = 16


COVERAGE_TYPES = OrderedDict({
    'buildings': {'id': COVT_BLD, 'desc': 'buildings'},
    'other': {'id': COVT_OTH, 'desc': 'other (typically appurtenant structures)'},
    'contents': {'id': COVT_CON, 'desc': 'contents'},
    'bi': {'id': COVT_BIT, 'desc': 'business interruption or other time-based coverage'},
    'pd': {'id': COVT_PDM, 'desc': 'property damage (buildings + other + contents)'},
    'all': {'id': COVT_ALL, 'desc': 'all (property damage + business interruption)'},
    'binp': {'id': COVT_BINP, 'desc': 'Business Interruption (non-property)'},
    'cbi': {'id': COVT_CBI, 'desc': 'Contingent Business Interruption'},
    'dias': {'id': COVT_DIAS, 'desc': 'Digital Assets (data, software and/or hardware recovery/replacement)'},
    'ext': {'id': COVT_EXT, 'desc': 'Extortion  (ransom amounts, negotiation fees, reward payments)'},
    'fin': {'id': COVT_FIN, 'desc': 'Financial Fraud - Theft of electronic funds, goods, services'},
    'inre': {'id': COVT_INRE, 'desc': 'Incident Response (IT Forensics, response obligations assessment, public relations, notifications and customer relations, monitoring services)'},
    'liab': {'id': COVT_LIAB, 'desc': 'Liability - 3rd party cyber liability (e.g. Network security, privacy, media) and directly associated costs (e.g legal defence costs)'},
    'reg': {'id': COVT_REG, 'desc': 'Regulatory and payment card industry (PCI) fines and proceedings'},
    'eno': {'id': COVT_ENO, 'desc': 'Errors & Omissions/Professional indemnity liability'},
    'cyb': {'id': COVT_CYB, 'desc': 'All Cyber coverages'}
})


SUPPORTED_COVERAGE_TYPES = OrderedDict({
    cov_type: cov_type_dict for cov_type, cov_type_dict in COVERAGE_TYPES.items()
    if cov_type in ['buildings', 'other', 'contents', 'bi', 'binp', 'cbi', 'dias', 'ext', 'fin', 'inre', 'liab', 'reg', 'eno']
})



import numba as nb
import numpy as np

from oasislmf.pytools.lec.utils import get_sample_idx_data, get_wheatsheaf_items_idx


@nb.njit(cache=True, error_model="numpy")
def fill_wheatsheaf_items(
    items,
    items_start_end,
    row_used_indices,
    outloss_vals,
    period_weights,
    max_summary_id,
    num_sidxs,
):
    # Track number of entries per summary_id
    summary_sidx_counts = np.zeros(max_summary_id * num_sidxs, dtype=np.int32)

    # First pass to count how many times each summary_id appears
    for idx in row_used_indices:
        summary_id, sidx, period_no = get_sample_idx_data(idx, max_summary_id, num_sidxs)
        summary_sidx_idx = get_wheatsheaf_items_idx(summary_id, sidx, num_sidxs)
        summary_sidx_counts[summary_sidx_idx] += 1

    # Compute cumulative start indices
    pos = 0
    for idx in range(max_summary_id * num_sidxs):
        if summary_sidx_counts[idx] > 0:
            items_start_end[idx][0] = pos  # Start index
            pos += summary_sidx_counts[idx]
            items_start_end[idx][1] = pos  # End index

    # Reset summary counts for inserting data
    summary_sidx_counts[:] = 0

    # Track which period_no are used if period_weights exists
    is_weighted = len(period_weights) > 0
    used_period_no = np.zeros(len(period_weights), dtype=np.bool_)

    # Second pass to populate the data array
    for idx in row_used_indices:
        summary_id, sidx, period_no = get_sample_idx_data(idx, max_summary_id, num_sidxs)
        summary_sidx_idx = get_wheatsheaf_items_idx(summary_id, sidx, num_sidxs)

        # Compute position in the flat array
        insert_idx = items_start_end[summary_sidx_idx][0] + summary_sidx_counts[summary_sidx_idx]

        # Store values
        items[insert_idx]["summary_id"] = summary_id
        items[insert_idx]["sidx"] = sidx
        items[insert_idx]["value"] = outloss_vals[idx]
        if is_weighted:
            # Fast lookup period_weights as they are numbered 1 to no_of_periods
            items[insert_idx]["period_weighting"] = period_weights[period_no - 1]["weighting"]
            items[insert_idx]["period_no"] = period_no
            used_period_no[period_no - 1] = True

        summary_sidx_counts[summary_sidx_idx] += 1

    return is_weighted, used_period_no



import numpy as np
import pandas as pd
from oasislmf.pytools.common.data import resolve_file


def read_csv_as_ndarray(stack, file_in, headers, dtype):
    file_in = resolve_file(file_in, "r", stack)

    csv_dtype = {key: col_dtype for key, (col_dtype, _) in dtype.fields.items()}
    try:
        df = pd.read_csv(file_in, delimiter=',', dtype=csv_dtype, usecols=list(csv_dtype.keys()))
    except pd.errors.EmptyDataError:
        return np.empty(0, dtype=dtype)

    data = np.empty(df.shape[0], dtype=dtype)
    for name in dtype.names:
        data[name] = df[name]
    return data



#!/usr/bin/env python

from contextlib import ExitStack
import logging
import sys
import numpy as np

from oasislmf.pytools.common.data import DEFAULT_BUFFER_SIZE, write_ndarray_to_fmt_csv
from oasislmf.pytools.converters.bintocsv.utils import (
    amplifications_tocsv,
    cdf_tocsv,
    complex_items_tocsv,
    coverages_tocsv,
    fm_tocsv,
    footprint_tocsv,
    gul_tocsv,
    lossfactors_tocsv,
    occurrence_tocsv,
    vulnerability_tocsv,
)
from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.data import TOOL_INFO

logger = logging.getLogger(__name__)


TOCSV_FUNC_MAP = {
    "amplifications": amplifications_tocsv,
    "cdf": cdf_tocsv,
    "complex_items": complex_items_tocsv,
    "coverages": coverages_tocsv,
    "fm": fm_tocsv,
    "footprint": footprint_tocsv,
    "gul": gul_tocsv,
    "lossfactors": lossfactors_tocsv,
    "occurrence": occurrence_tocsv,
    "vulnerability": vulnerability_tocsv,
}


def default_tocsv(stack, file_in, file_out, file_type, noheader):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    file_in = resolve_file(file_in, "rb", stack)
    if file_in == sys.stdin.buffer:
        data = np.frombuffer(file_in.read(), dtype=dtype)
    else:
        data = np.fromfile(file_in, dtype=dtype)
    num_rows = data.shape[0]

    if not noheader:
        file_out.write(",".join(headers) + "\n")

    buffer_size = DEFAULT_BUFFER_SIZE
    for start in range(0, num_rows, buffer_size):
        end = min(start + buffer_size, num_rows)
        buffer_data = data[start:end]
        write_ndarray_to_fmt_csv(file_out, buffer_data, headers, fmt)


def bintocsv(file_in, file_out, file_type, noheader=False, **kwargs):
    """Convert bin file to csv file based on file type
    Args:
        file_in (str | os.PathLike): Input file path
        file_out (str | os.PathLike): Output file path
        file_type (str): File type str from SUPPORTED_BINTOCSV
        noheader (bool): Bool to not output header. Defaults to False.
    """
    with ExitStack() as stack:
        file_out = resolve_file(file_out, "w", stack)

        tocsv_func = TOCSV_FUNC_MAP.get(file_type, default_tocsv)
        tocsv_func(stack, file_in, file_out, file_type, noheader, **kwargs)



__all__ = [
    "amplifications_tobin",
    "complex_items_tobin",
    "coverages_tobin",
    "damagebin_tobin",
    "fm_tobin",
    "footprint_tobin",
    "gul_tobin",
    "lossfactors_tobin",
    "occurrence_tobin",
    "returnperiods_tobin",
    "summarycalc_tobin",
    "vulnerability_tobin",
]

from .amplifications import amplifications_tobin
from .complex_items import complex_items_tobin
from .coverages import coverages_tobin
from .damagebin import damagebin_tobin
from .fm import fm_tobin
from .footprint import footprint_tobin
from .gul import gul_tobin
from .lossfactors import lossfactors_tobin
from .occurrence import occurrence_tobin
from .returnperiods import returnperiods_tobin
from .summarycalc import summarycalc_tobin
from .vulnerability import vulnerability_tobin



# evepy/manager.py

from contextlib import ExitStack
import logging
from pathlib import Path
import numpy as np

from oasislmf.pytools.utils import redirect_logging
from oasislmf.pytools.common.data import oasis_int, resolve_file

logger = logging.getLogger(__name__)

DEFAULT_EVENTS_FILE = Path('input/events.bin')
NUMPY_RANDOM_SEED = 723706


def read_events(input_file):
    """Read the event IDs from the binary events file.

    Args:
        input_file (str | os.PathLike): Path to binary events file.
    """
    return np.fromfile(input_file, dtype=oasis_int)


def stream_events(events, stream_out):
    """Stream the output events.

    Args:
        events (Iterable): Iterable containing the events to stream.
        stream_out (File object): File object with `write` method for handling output.
    """
    for e in events:
        stream_out.write(np.int32(e).tobytes())


def calculate_events_per_process(n_events, total_processes):
    """Calculate number of events per process.
    """
    events_per_process, remainder = divmod(n_events, total_processes)
    return events_per_process + bool(remainder)  # add 1 if remainder


def partition_events__no_shuffle(events, process_number, total_processes):
    """Assign events in the order they are loaded to each process in turn. Only
    output the event IDs allocated to the given `process_number`.

    Args:
        events (np.array): Array of ordered event IDs.
        process_number (int): The process number to receive a partition of events.
        total_processes (int): Total number of processes to distribute the events over.
    """
    events_per_process = calculate_events_per_process(len(events), total_processes)
    return events[(process_number - 1) * events_per_process:
                  process_number * events_per_process]


def partition_events__random_builtin(events, process_number, total_processes):
    """Shuffle the events randomly and allocate to each process using builtin
    shuffle. Only output the event IDs to the given `process_number`.

    Note that this can be memory intensive. For `len(events) > 10**5` recommend
    using `partition_events__random`.

    Args:
        events (np.array): Array of ordered event IDs.
        process_number (int): The process number to receive a partition of events.
        total_processes (int): Total number of processes to distribute the events over.
    """
    rng = np.random.default_rng(NUMPY_RANDOM_SEED)
    rng.shuffle(events)
    return partition_events__no_shuffle(events, process_number, total_processes)


def partition_events__random(events, process_number, total_processes):
    """Shuffle the events randomly and allocate to each process. Only output
    the event IDs to the given `process_number`. Generates an iterator.

    Randomisation is implemented using the Fisher-Yates algorithm.

    Args:
        events (np.array): Array of ordered event IDs.
        process_number (int): The process number to receive a partition of events.
        total_processes (int): Total number of processes to distribute the events over.
    """
    rng = np.random.default_rng(NUMPY_RANDOM_SEED)

    for i in range(len(events) - 1, 0, -1):
        j = rng.integers(0, i + 1)

        events[i], events[j] = events[j], events[i]

        if (i - process_number) % total_processes == 0:
            yield events[i]

    if process_number % total_processes == 0:  # Event at 0 index
        yield events[0]


def partition_events__round_robin(events, process_number, total_processes):
    """Partition the events sequentially in a round robin style per process.
    Only output the events allocated to the given `process_number`.

    Args:
        events (np.array): Array of ordered event IDs.
        process_number (int): The process number to receive a partition of events.
        total_processes (int): Total number of processes to distribute the events over.
    """
    return events[np.arange(process_number - 1, len(events), total_processes)]


def run(input_file, process_number, total_processes, no_shuffle=False,
        randomise=False, randomise_builtin=False, output_file='-',
        ):
    """Generate event ID partitions as a binary data stream with shuffling. By
    default the events are shuffled by assiging to processes one by one
    cyclically.

    Args:
        input_file (str | os.PathLike): Path to binary events file. If None
        then defaults to DEFAULT_EVENTS_FILE.
        process_number (int): The process number to receive a partition of events.
        total_processes (int): Total number of processes to distribute the events over.
        no_shuffle (bool, optional): Disable shuffling events. Events are split
            and distributed into blocks in the order they are input. Takes
            priority over `randomise(_builtin)`.
        randomise (bool, optional): Shuffle events randomly in the blocks and
            stream events on the fly. If `no_shuffle` is `True` then it takes
            priority.
        randomise_builtin (bool, optional): Shuffle events randomly in the blocks using
            builtin shuffle. If `no_shuffle` or `randomise` is `True` then it
            they take priority.
        output_file (str | os.PathLike): Path to output file. If '-' then outputs to stdout.
    """
    if input_file is None:
        input_file = DEFAULT_EVENTS_FILE

    # Check input file is valid
    input_file = Path(input_file)
    if not input_file.exists():
        raise FileNotFoundError(f"ERROR: File \'{input_file}\' does not exist.")
    if not input_file.is_file():
        raise ValueError(f"ERROR: \'{input_file}\' is not a file.")

    # Check shuffle and randomise settings
    if no_shuffle and (randomise or randomise_builtin):
        logger.warning("Warning: `no_shuffle` and `randomise(_builtin)` options are incompatible. Ignoring `randomise(_builtin)`.")
        randomise = False
        randomise_builtin = False

    if randomise and randomise_builtin:
        logger.warning("Warning: `randomise` and `randomise_builtin` options are incompatible. Ignoring `randomise_builtin`.")
        randomise_builtin = False

    events = read_events(input_file)

    if no_shuffle:
        event_partitions = partition_events__no_shuffle(events,
                                                        process_number,
                                                        total_processes)
    elif randomise:
        event_partitions = partition_events__random(events,
                                                    process_number,
                                                    total_processes)
    elif randomise_builtin:
        event_partitions = partition_events__random_builtin(events,
                                                            process_number,
                                                            total_processes)
    else:
        event_partitions = partition_events__round_robin(events,
                                                         process_number,
                                                         total_processes)

    with ExitStack() as stack:
        stream_out = resolve_file(output_file, 'wb', stack)

        stream_events(event_partitions, stream_out)


@redirect_logging(exec_name='evepy')
def main(input_file=None, process_number=None, total_processes=None,
         no_shuffle=False, randomise=False, randomise_builtin=False, output_file='-', **kwargs):

    run(input_file=input_file, process_number=process_number,
        total_processes=total_processes, no_shuffle=no_shuffle,
        randomise=randomise, randomise_builtin=randomise_builtin, output_file=output_file)



import csv
import msgpack
import numpy as np
from oasislmf.pytools.common.event_stream import mv_read
from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.data import TOOL_INFO


def complex_items_tocsv(stack, file_in, file_out, file_type, noheader):
    header_dtype = TOOL_INFO[file_type]["dtype"]

    file_in = resolve_file(file_in, "rb", stack)
    byte_data = np.frombuffer(file_in.read(), dtype=np.uint8)

    writer = csv.writer(file_out)

    # Write header line manually
    if not noheader:
        writer.writerow(("item_id", "coverage_id", "model_data", "group_id"))

    cursor = 0
    while cursor < byte_data.size:
        header_record = np.zeros((), dtype=header_dtype)
        header_record["item_id"], cursor = mv_read(byte_data, cursor, header_dtype["item_id"], header_dtype["item_id"].itemsize)
        header_record["coverage_id"], cursor = mv_read(byte_data, cursor, header_dtype["coverage_id"], header_dtype["coverage_id"].itemsize)
        header_record["group_id"], cursor = mv_read(byte_data, cursor, header_dtype["group_id"], header_dtype["group_id"].itemsize)
        header_record["model_data_len"], cursor = mv_read(byte_data, cursor, header_dtype["model_data_len"], header_dtype["model_data_len"].itemsize)

        model_data_len = header_record["model_data_len"]
        model_data_bytes = byte_data[cursor:cursor + model_data_len].tobytes()
        cursor += model_data_len

        # Unpack msgpack
        if msgpack.version >= (1, 0, 0):
            model_data = msgpack.unpackb(model_data_bytes, raw=False)
        else:
            model_data = msgpack.unpackb(model_data_bytes)
            if isinstance(model_data, bytes):
                model_data = model_data.decode("utf-8")

        writer.writerow((
            header_record["item_id"],
            header_record["coverage_id"],
            model_data,
            header_record["group_id"]
        ))



from oasislmf.pytools.common.data import fm_programme_headers, fm_programme_dtype, fm_programme_fmt


headers = fm_programme_headers
dtype = fm_programme_dtype
fmt = fm_programme_fmt
cli_support = ['bintocsv', 'csvtobin']



import zlib
import numpy as np
import pandas as pd
import re
import sys

from oasislmf.pytools.common.data import resolve_file, write_ndarray_to_fmt_csv
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.getmodel.common import Event, EventIndexBin, EventIndexBinZ, FootprintHeader


def _check_event_from_to(event_from_to):
    from_event = -1
    to_event = -1
    if event_from_to == None:
        return True, from_event, to_event

    regex_match = re.fullmatch(r'(\d+)-(\d+)', event_from_to)
    if not regex_match:
        raise ValueError(f"Invalid format for event_from_to string: {event_from_to}. String must be of format \"[int1]-[int2]\"")

    from_event, to_event = map(int, regex_match.groups())
    if from_event > to_event:
        raise ValueError(f"Invalid event range: {from_event} > {to_event}")

    return False, from_event, to_event


def _read_footprint_zips(stack, file_in, idx_file_in):
    footprint_file = resolve_file(file_in, mode="rb", stack=stack)

    if footprint_file == sys.stdin.buffer:
        footprint = np.frombuffer(footprint_file.read(), dtype="u1")
    else:
        footprint = np.fromfile(footprint_file, dtype="u1")

    footprint_header = np.frombuffer(footprint[:FootprintHeader.size].tobytes(), dtype=FootprintHeader)

    uncompressedMask = 1 << 1
    uncompressed_size = int(footprint_header['has_intensity_uncertainty'] & uncompressedMask)

    if uncompressed_size:
        index_dtype = EventIndexBinZ
    else:
        index_dtype = EventIndexBin

    footprint_index_file = resolve_file(idx_file_in, mode="rb", stack=stack)
    footprint_mmap = np.memmap(footprint_index_file, dtype=index_dtype, mode='r')

    footprint_index = pd.DataFrame(footprint_mmap, columns=footprint_mmap.dtype.names).set_index('event_id').to_dict('index')

    return footprint, footprint_index


def _read_footprint_bins(stack, file_in, idx_file_in):
    footprint_file = resolve_file(file_in, mode="rb", stack=stack)

    if footprint_file == sys.stdin.buffer:
        footprint = np.frombuffer(footprint_file.read(), dtype="u1")
    else:
        footprint = np.fromfile(footprint_file, dtype="u1")

    footprint_index_file = resolve_file(idx_file_in, mode="rb", stack=stack)
    footprint_mmap = np.memmap(footprint_index_file, dtype=EventIndexBin, mode='r')

    footprint_index = pd.DataFrame(footprint_mmap, columns=footprint_mmap.dtype.names).set_index('event_id').to_dict('index')

    return footprint, footprint_index


def footprint_tocsv(stack, file_in, file_out, file_type, noheader, idx_file_in, zip_files, event_from_to):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]
    no_event_range, from_event, to_event = _check_event_from_to(event_from_to)

    if zip_files:
        footprint, footprint_index = _read_footprint_zips(stack, file_in, idx_file_in)
    else:
        footprint, footprint_index = _read_footprint_bins(stack, file_in, idx_file_in)

    if not noheader:
        file_out.write(",".join(headers) + "\n")

    event_ids = sorted(footprint_index.keys())
    for event_id in event_ids:
        if no_event_range or (event_id >= from_event and event_id <= to_event):
            offset = footprint_index[event_id]["offset"]
            size = footprint_index[event_id]["size"]
            if zip_files:
                event_data = zlib.decompress(footprint[offset:offset + size].tobytes())
                event_data = np.frombuffer(event_data, dtype=Event)
            else:
                event_data = np.frombuffer(footprint[offset:offset + size].tobytes(), dtype=Event)
            event_csv_data = np.empty(len(event_data), dtype=dtype)
            event_csv_data["event_id"] = event_id
            event_csv_data["areaperil_id"] = event_data["areaperil_id"]
            event_csv_data["intensity_bin_id"] = event_data["intensity_bin_id"]
            event_csv_data["probability"] = event_data["probability"]
            write_ndarray_to_fmt_csv(file_out, event_csv_data, headers, fmt)



from abc import ABC, abstractmethod

from oasislmf.utils.documentation.markdown import MarkdownGenerator


class BaseJsonToMarkdownGenerator(ABC):
    """
    Base JSON to Markdown Generator class
    """

    def __init__(self, full_schema, data_path, doc_out_dir, markdown_generator=None):
        """
        Args:
            full_schema (Dict): Full schema file as dictionary
            data_path (str | os.PathLike): Path to data folder for any relative file paths
            doc_out_dir (str | os.PathLike): Path to documentation file output folder for any relative file paths
            markdown_generator (MarkdownGenerator, optional): MarkdownGenerator class. Defaults to None.
        """
        self.full_schema = full_schema
        self.data_path = data_path
        self.doc_out_dir = doc_out_dir
        self.md = markdown_generator
        if not markdown_generator:
            self.md = MarkdownGenerator()

    def _resolve_internal_ref(self, ref):
        """Resolves a $ref in the schema (only internal refs supported).
        Args:
            ref (str): Reference string of format #/$<reftitle>/<refname>
        Returns:
            ref_schema (Dict): Data Properties from reference schema as dictionary
        """
        parts = ref.strip("#/").split("/")
        ref_schema = self.full_schema
        for part in parts:
            ref_schema = ref_schema.get(part, {})
        return ref_schema

    @abstractmethod
    def generate(self, json_data, generate_toc=False):
        """Top level function to process entire dict to markdown text
        Args:
            json_data (Dict): Json data as dictionary
            generate_toc (bool, Optional): Generate table of contents bool. Defaults to False.
        Returns:
            markdown_txt (str): Markdown text
        """
        pass
